import numpy as np
import sympy as sp

from cge_modeling.tools.numba_tools import numba_lambdify


def test_numba_lambdify_single_output():
    a, b, c = sp.symbols("a b c")
    f = a * b + c
    f_func = numba_lambdify(exog_vars=[a, b, c], expr=f)

    assert f_func([1, 2, 3]) == 5


def test_numba_lambdify_multi_output():
    a, b, c = sp.symbols("a b c")
    f1 = a * b + c
    f2 = a + b + c

    f_func = numba_lambdify(exog_vars=[a, b, c], expr=[f1, f2], ravel_outputs=True)

    outputs = f_func([1, 2, 3])
    np.testing.assert_allclose(outputs, [np.array([5]), np.array([6])])


def test_numba_lambdify_matrix_output():
    a, b, c = sp.symbols("a b c")
    f1 = sp.Matrix([a * b + c, a + b + c])

    f_func = numba_lambdify(exog_vars=[a, b, c], expr=f1)

    outputs = f_func([1, 2, 3])
    np.testing.assert_allclose(outputs, np.array([5, 6]).reshape((2, 1)))



import numpy as np
import pytensor.tensor as pt
import pytest

from cge_modeling.base.primitives import Parameter, Variable
from cge_modeling.pytensorf.compile import (
    compile_cge_model_to_pytensor,
    compile_euler_approximation_function,
    object_to_pytensor,
)
from cge_modeling.tools.pytensor_tools import make_jacobian
from tests.utilities.models import load_model_1, load_model_2

test_cases = [
    (Parameter, "K_d", {}),
    (Variable, "Y", {"t": [0, 1, 2]}),
    (Variable, "Y", {"t": [0, 1, 2], "s": [0, 1, 2]}),
]


@pytest.mark.parametrize(
    "cls, name, coords", test_cases, ids=["Parameter", "Variable", "Variable2d"]
)
def test_object_to_pytensor(cls, name, coords):
    # noinspection PyArgumentList
    obj = cls(
        name=name,
        dims=list(coords.keys()),
        description="A lovely item from group <dim:i>",
    )

    pt_obj = object_to_pytensor(obj, coords)
    assert pt_obj.name == name
    assert pt_obj.ndim == len(coords)
    assert pt_obj.type.shape == tuple(len(coords[dim]) for dim in coords.keys())


@pytest.mark.parametrize(
    "model_func, kwargs",
    [
        (
            load_model_1,
            {"backend": "pytensor", "compile": False, "parse_equations_to_sympy": True},
        ),
        (
            load_model_2,
            {
                "backend": "pytensor",
                "compile": False,
                "parse_equations_to_sympy": False,
            },
        ),
    ],
    ids=["1_Sector", "3_Sector"],
)
def test_compile_to_pytensor(model_func, kwargs):
    cge_model = model_func(**kwargs)
    n_variables = n_eq = len(cge_model.unpacked_variable_names)
    n_params = len(cge_model.unpacked_parameter_names)

    (variables, parameters), (model, jac, B) = compile_cge_model_to_pytensor(cge_model)

    assert model.type.shape == (n_eq,)
    assert jac.type.shape == (n_eq, n_variables)
    assert B.type.shape == (n_eq, n_params)


def test_compile_euler_approximation_function():
    # Example problem from Notes and Problems from Applied General Equilibrium Economics, Chapter 3

    variables = v1, v2 = [pt.dscalar(name) for name in ["v1", "v2"]]
    parameters = v3 = pt.dscalar("v3")
    equations = pt.stack([v1**2 * v3 - 1, v1 + v2 - 2])

    def f_analytic(v3):
        v1 = 1 / np.sqrt(v3)
        v2 = 2 - v1
        return np.array([v1, v2])

    mode = "FAST_COMPILE"

    A = make_jacobian(equations, variables)
    B = make_jacobian(equations, [parameters])
    f_1 = compile_euler_approximation_function(
        A=A,
        B=B,
        variables=variables,
        parameters=[parameters],
        n_steps=1,
        mode=mode,
    )
    f_10 = compile_euler_approximation_function(
        A=A,
        B=B,
        variables=variables,
        parameters=[parameters],
        n_steps=10,
        mode=mode,
    )
    f_100 = compile_euler_approximation_function(
        A=A,
        B=B,
        variables=variables,
        parameters=[parameters],
        n_steps=100,
        mode=mode,
    )
    f_10k = compile_euler_approximation_function(
        A=A,
        B=B,
        variables=variables,
        parameters=[parameters],
        n_steps=10_000,
        mode=mode,
    )

    initial_point = [1, 1]
    v3_initial = 1
    v3_final = 2

    analytic_solution = f_analytic(v3_final)
    approximate_solutions = [
        np.c_[f(*initial_point, v3_initial, np.array([v3_final]))[:2]].T
        for f in [f_1, f_10, f_100, f_10k]
    ]
    errors = np.c_[[solution[-1] - analytic_solution for solution in approximate_solutions]]

    # Test the errors are monotonically decreasing in the number of steps
    assert np.all(np.diff(np.abs(errors), axis=0) < 0)

    # Test that the solution is close to the analytic solution at 10,000 steps
    assert np.allclose(approximate_solutions[-1][-1, :], analytic_solution, atol=1e-5)



import string

from collections.abc import Sequence
from itertools import combinations, product
from string import Template
from typing import Literal, cast

from cge_modeling.tools.pytensor_tools import at_least_list

BACKEND_TYPE = Literal["numba", "pytensor"]


def _check_pairwise_lengths_match(names, args):
    list_args = [at_least_list(arg) for arg in args]
    arg_pairs = combinations(list_args, 2)
    name_pairs = combinations(names, 2)
    for name_pair, arg_pair in zip(name_pairs, arg_pairs):
        if len(arg_pair[0]) != len(arg_pair[1]):
            raise ValueError(f"Lengths of {name_pair[0]} and {name_pair[1]} do not match")


def unwrap_singleton_list(x):
    if isinstance(x, list) and (len(x) == 1):
        return x[0]
    else:
        return x


def unpack_string_inputs(
    *inputs: str | list[str, ...],
    dims: Sequence[str] | None = None,
    coords: dict[str, list[str, ...]] | None = None,
) -> list[list[str, ...]]:
    """
    Unpack string inputs into lists of variables using the provided dims and coords.


    Parameters
    ----------
    inputs: str or list of str
        Inputs to be unpacked. If an input is a string, it will be unpacked into N inputs using the provided dims.

    dims: Sequence of str or str
        Sequence of named dimensions for variables in the equations to be generated. All dimensions should appear in
        the coords dictionary as keys, otherwise an error will be raised.

    coords: dict of str: list of str
        Dictionary of coordinates for the model, mapping dimension names to lists of labels.

    Returns
    -------
    unpacked_inputs: list of lists of str
        unpacked factors and factor prices

    Notes
    -----
    If inputs are lists, this function will return them unchanged.
    """
    inputs = list(map(unwrap_singleton_list, inputs))

    if all([isinstance(x, list) for x in inputs]):
        return list(map(at_least_list, inputs))

    if coords is None or dims is None:
        return list(map(at_least_list, inputs))

    if not all([key in coords.keys() for key in dims]):
        raise KeyError(f"Dimensions {dims} not found in coords")

    labels = [coords[dim] for dim in dims]
    label_prod = product(*labels)
    outputs: list[list[str, ...]] = [[] for _ in inputs]

    for labels in label_prod:
        label_list = [x for x in labels]

        for i, input in enumerate(inputs):
            outputs[i].append("_".join([input, *label_list]))

    return outputs


def _add_second_alpha(alpha: str | list[str, ...], factors: str | list[str, ...]) -> list[str, ...]:
    if not isinstance(factors, list) or len(factors) != 2:
        # If there are not exactly two factors, we don't need to add a second alpha. It will either be unpacked
        # later, or it will be an error.
        return cast(list[str, ...], at_least_list(alpha))

    alpha = cast(list[str, ...], at_least_list(alpha))

    if len(factors) == 2 and len(alpha) == 1:
        alpha.append(f"1 - {alpha[0]}")
        return alpha

    return alpha


def cobb_douglass(
    factors: list[str, ...],
    factor_prices: list[str, str],
    output: str,
    output_price: str,
    factor_shares: str,
    TFP: str = "1",
    use_value_definition: bool = False,
) -> tuple[str, ...]:
    """
    Generate string equations representing a Cobb-Douglas production process.

    A Cobb-Douglas production process is defined as:

    .. math::
        Y = A \\prod_{i \\in I} X_i^{\\alpha_i}

    Where :math:`A` is the total factor productivity parameter, :math:`\\alpha_i` is the share of factor :math:`X_i` in
    the production process. The profit maximization problem for this production function generates the following factor
    demands:

    .. math::
        X_i = \frac{Y}{A} \frac{\\alpha_i P_Y A}{P_i}

    A Cobb-Douglas can be shown to be a special case of the CES production function, when the elasticity of substitution
    between input factors is :math:`\\epsilon = 1`. In this special case, the ratio of factor inputs ..math:`\\frac{X_i}{X_j}`
    is always equal to the ratio of factor prices ..math:`\\frac{P_i}{P_j}`. That is, the proportion of the i-th
    input factor used in the production process is always exactly ..math:`\\alpha_i`, and prices adjust to ensure that
    this is so.

    Parameters
    ----------
    factors: list of str
        Production factors

    factor_prices: list of str
        Production factor prices. Must be the same length as factors.

    output: str
        Name of the output of the production function.

    output_price: str
        Name of the price of the output.

    factor_shares: str or list of str
        Factor shares in the production function. If a list, it should be the same length as factors, or one less. If
        one less, the last factor share is assumed to be 1 - sum(factor_shares).

        A string is valid only if len(factors) == 2, in which case the string is interpreted as [factor_shares].

    TFP: str, default "1"
        Technology parameter

    Returns
    -------
    production_function: str
        Equation representing the Cobb-Douglas production function
    factor_demands: list of str
        A list of equations, one for each factor, representing the factor demands
    """
    if isinstance(factor_shares, str):
        factor_shares = [factor_shares]

    factor_shares = _add_second_alpha(factor_shares, factors)
    _check_pairwise_lengths_match(
        ["factors", "factor_prices", "factor_shares"],
        [factors, factor_prices, factor_shares],
    )

    if not use_value_definition:
        production_inner_template = Template("$factor ** ($factor_share)")
        production_inner = [
            production_inner_template.safe_substitute(factor=factor, factor_share=factor_share)
            for factor, factor_share in zip(factors, factor_shares)
        ]

        eq_production = Template(
            "$output = $TFP * " + " * ".join(production_inner)
        ).safe_substitute(output=output, TFP=TFP)
    else:
        total_factor_value = " + ".join(
            [f"{factor_price} * {factor}" for factor_price, factor in zip(factor_prices, factors)]
        )
        eq_production = Template(
            "$output * $output_price = $total_factor_value",
        ).safe_substitute(
            output=output,
            output_price=output_price,
            total_factor_value=total_factor_value,
        )

    factor_demand_template = Template(
        "$factor = ($factor_share) * $output * ($output_price) / ($factor_price)"
    )

    eq_fac_demands = [
        factor_demand_template.safe_substitute(
            factor=factor,
            factor_price=factor_price,
            output=output,
            output_price=output_price,
            factor_share=factor_share,
            TFP=TFP,
        )
        for factor, factor_price, factor_share in zip(factors, factor_prices, factor_shares)
    ]

    return (eq_production, *tuple(eq_fac_demands))


def CES(
    factors: list[str, ...] | str,
    factor_prices: list[str, ...] | str,
    output: str,
    output_price: str,
    TFP: str,
    factor_shares: str | list[str, ...],
    epsilon: str,
    expand_price_dim: Literal["input", "output", "both", None] = None,
    use_value_definition=False,
    *args,
    **kwargs,
) -> tuple[str, ...]:
    """
    Generate string equations representing a CES production process.

    A CES production process is defined as:

    .. math::
        Y = A \\left( \\sum_{i \\in I} \alpha_i X_i^{\frac{\\epsilon - 1}{\\epsilon}} \right)^{\frac{\\epsilon}{\\epsilon - 1}}

    Where :math:`A` is the total factor productivity parameter, :math:`\\alpha_i` is the share of factor :math:`X_i` in
    the production process, and :math:`\\epsilon` is the elasticity of substitution parameter. The profit maximization
    problem for this production function generates the following factor demands:

    .. math::
        X_i = \frac{Y}{A} \\left( \frac{\alpha_i P_Y A}{P_i} \right)^{\\epsilon}

    This definition is the one most frequently used in the CGE literature.

    Parameters
    ----------
    factors: str or list of str
        Production factors. If a string is provided, it will be unpacked into N prices using the
        provided dims.

    factor_prices: str or list of str
        Production factor prices. If a string is provided, it will be unpacked into N prices using the
        provided dims.

    output: str
        Name of the output of the production function

    output_price: str
        Name of the price of the output

    TFP: str
        Technology parameter

    factor_shares: str, or list of str
        Factor shares in the production function. If a string is provided and len(factors) == 2, the second factor
        share will be set to (1 - alpha). If a list of strings is provided, it must be the same length as factors.

    epsilon:
        elasticity parameter

    expand_price_dim: str, one of 'input', 'output', 'both', or None
        Price dimension to expand. If expanded, a price P will be printed as P[:, None].

    args, kwargs:
        Ignored; included for signature compatibility with other production functions

    Returns
    -------
    (variables, parameters, equations): tuple of tuples of strings
        output and factor equations
    """

    factor_shares = _add_second_alpha(factor_shares, factors)
    _check_pairwise_lengths_match(
        ["factors", "factor_prices", "factor_shares"],
        [factors, factor_prices, factor_shares],
    )

    if not use_value_definition:
        production_inner_template = Template(
            "($factor_shares) * $factor ** (($epsilon - 1) / $epsilon)"
        )

        production_inner = [
            production_inner_template.safe_substitute(
                factor=factor, factor_shares=factor_share, epsilon=epsilon
            )
            for factor, factor_share in zip(factors, factor_shares)
        ]

        eq_production = Template(
            "$output = $TFP * ($inner) ** ($epsilon / ($epsilon - 1))"
        ).safe_substitute(
            output=output, inner=" + ".join(production_inner), TFP=TFP, epsilon=epsilon
        )
    else:
        total_factor_value = " + ".join(
            [f"{factor_price} * {factor}" for factor_price, factor in zip(factor_prices, factors)]
        )
        eq_production = Template(
            "$output * $output_price = $total_factor_value",
        ).safe_substitute(
            output=output,
            output_price=output_price,
            total_factor_value=total_factor_value,
        )

    factor_demand_template = Template(
        "$factor = $output / $TFP * (($factor_share) * $output_price * $TFP / ($factor_price)) ** "
        "$epsilon"
    )

    if expand_price_dim in ["output", "both"]:
        output_price = f"{output_price}[:, None]"
    if expand_price_dim in ["input", "both"]:
        factor_prices = [f"{factor_price}[:, None]" for factor_price in factor_prices]

    eq_fac_demands = [
        factor_demand_template.safe_substitute(
            factor=factor,
            factor_price=factor_price,
            output=output,
            output_price=output_price,
            factor_share=factor_share,
            TFP=TFP,
            epsilon=epsilon,
        )
        for factor, factor_price, factor_share in zip(factors, factor_prices, factor_shares)
    ]

    return (eq_production, *tuple(eq_fac_demands))


def dixit_stiglitz(
    factors: str | list[str, ...],
    factor_prices: str | list[str, ...],
    output: str,
    output_price: str,
    epsilon: str,
    dims: str,
    coords: dict[str, list[str, ...]],
    backend: BACKEND_TYPE = "numba",
    TFP: str | None = None,
    factor_shares: str | None = None,
    use_value_definition=True,
) -> tuple[str, str]:
    """
    Generate string equations representing a Dixit-Stiglitz production process.

    Here, the Dixit-Stiglitz production process is defined as:

    .. math::
        Y = A \\left( \\sum_{i \\in I} \alpha_i X_i^{\frac{\\epsilon - 1}{\\epsilon}} \right)^{\frac{\\epsilon}{\\epsilon - 1}}

    Which generates the following factor demands:

    .. math::
        X_i = \frac{Y}{A} \\left( \frac{\alpha_i P_Y A}{P_i} \right)^{\\epsilon}

    Many researchers omit either the technology parameter, the factor shares, or both. If the technology parameter is
    not provided (i.e. :math:`A` is None), it is assumed to be 1. If the factor shares are omitted they are also set
    to 1.

    Parameters
    ----------
    factors: str or list of str
        Production factor. If a list is provided, it must be of length 1.

        .. warning::
        Despite the plural name, the Dixit-Stiglitz production function expects only a single facto input. The name
        is plural to be consistent with other production functions.

    factor_prices: str or list of str
        Production factor prices. If a list is provided, it must be of length 1.

    output: str
        Name of the output of the production function

    output_price: str
        Name of the price of the output

    TFP: str, optional
        Technology parameter. If not provided, it is omitted from the equations, which is equivalent to setting it to 1.

    factor_shares: str, optional
        Factor shares in the production function. If a list is provided, it must be of length 1. If not provided, it is
        omitted from the equations, which is equivalent to setting it to 1.

    epsilon:
        Elasticity of substitution parameter

    dims: Sequence of str or str
        Sequence of named dimensions for variables in the equations to be generated. All dimensions should appear in
        the coords dictionary as keys, otherwise an error will be raised.

    coords: dict of str: list of str
        Dictionary of coordinates for the model, mapping dimension names to lists of labels.

    backend: str, one of 'numba' or 'pytensor'
        Backend that will parse the equations. Only relevant for aggregation functions like Sum and Prod. Elementwise
        operations are interoperable between backends.

    Returns
    -------
    (variables, parameters, equations): tuple of tuples of strings
        output and factor equations
    """

    for var, name in zip(
        [factors, factor_prices, factor_shares],
        ["factors", "factor_prices", "factor_shares"],
    ):
        if isinstance(var, list) and len(var) != 1:
            raise ValueError(
                f"Dixit-Stiglitz production function expects only a single factor input, but provided "
                f"{name} has length {len(var)}"
            )

    factors, factor_prices, factor_shares = map(
        unwrap_singleton_list, (factors, factor_prices, factor_shares)
    )

    share_str = "$factor_share * " if factor_shares is not None else ""
    TFP_str = "$TFP * " if TFP is not None else ""

    kernel_str = f"{share_str}$factor ** (($epsilon - 1) / $epsilon)"

    kernel_template = Template(kernel_str)
    kernel = kernel_template.safe_substitute(
        factor_share=factor_shares, factor=factors, epsilon=epsilon
    )

    dim_len = len(coords[dims]) - 1

    if backend == "numba":
        rhs_str = "Sum($kernel, ($dims, 0, $dim_len)) ** ($epsilon / ($epsilon - 1))"
    elif backend == "pytensor":
        rhs_str = "($kernel).sum() ** ($epsilon / ($epsilon - 1))"
    else:
        raise ValueError(f"backend must be one of 'numba' or 'pytensor', found {backend}")

    if not use_value_definition:
        production_function = Template(f"$output = {TFP_str}{rhs_str}").safe_substitute(
            output=output,
            TFP=TFP,
            kernel=kernel,
            epsilon=epsilon,
            dims=dims,
            dim_len=dim_len,
        )
    else:
        production_function = Template(
            "$output = ($factor_prices * $factors).sum() / ($output_price)",
        ).safe_substitute(
            output=output,
            output_price=output_price,
            factor_prices=factor_prices,
            factors=factors,
        )

    demand_template = f"$factor = $output / {TFP_str}({TFP_str}{share_str}$output_price / $factor_price) ** $epsilon"

    factor_demands = Template(demand_template).safe_substitute(
        factor=factors,
        output=output,
        TFP=TFP,
        factor_share=factor_shares,
        output_price=output_price,
        factor_price=factor_prices,
        epsilon=epsilon,
    )

    return production_function, factor_demands


def _1d_leontief(
    factors: str | list[str, ...],
    factor_prices: str | list[str, ...],
    output: str,
    output_price: str,
    factor_shares: str | list[str, ...],
    *args,
    **kwargs,
) -> tuple[str, ...]:
    """
    Helper function to generate Leontief production equaitons for the case where ndim == 1.

    """
    zp_rhs = " + ".join([f"{price} * {factor}" for factor, price in zip(factors, factor_prices)])
    zero_profit = f"{output} = ({zp_rhs}) / ({output_price})"

    factor_demands = (
        f"{factor} = {output} * {share}" for factor, share in zip(factors, factor_shares)
    )

    return (zero_profit, *tuple(factor_demands))


def _2d_leontief(
    factors: str | list[str, ...],
    factor_prices: str | list[str, ...],
    output: str,
    output_price: str,
    factor_shares: str | list[str, ...],
    dims: str,
    coords: dict[str, list[str, ...]],
    sum_dim: str | None = None,
    expand_price_dim: bool = True,
    transpose_output: bool = True,
    backend: Literal["numba", "pytensor"] = "numba",
) -> tuple[str, ...]:
    r"""
    Helper function to generate Leontief production equations for the case where ndim == 2.

    Notes
    -----
    Sympy is quite restrictive in terms of broadcasting, so there are many hoops to jump through here.
    Assume that X is an N x N matrix of factor demands, phi is an N x N matrix of technological coefficients,
    Y is an N x 1 vector of outputs and P_Y is a vectors of prices Y. P_X is either a vector or a matrix of factor
    prices, depending on the situaiton. If it is a vector, set expand_price_dim to True; if it is a matrix, set it to
    False.

    The important thing about X is that it is indexed by the same labels twice. The columns represent demands by
    the j-th label to the i-th label. The rows, on the other hand, represent supply by the  i-th label to the
    j-th label. This is the standard representation of an input-output matrix in CGE modeling.

    The zero profit condition will ask us to sum across the **rows** of X:

    [X_00,   | , X_02, X_03]
    [X_10,   | , X_12, X_13]
    [--------+-------------]-> This sum gives total demand for goods produced by the i-th label
    [X_20,   | , X_22, X_23]
            \_/
             This sum gives total supply of goods produced by the j-th label

    If broadcasting, this operation is quite simple: we multiply X * phi, then sum across the columns with axis=0.
    With simpy indexed symbols, however, this is not possible. First, these don't have a notion of axis, so we
    have to manually manipulate the indices to get what we want.

    - First, make P_Y * Y a "column vector" by swapping the core and batch dimensions
    - Next, "transpose" X by swapping the core and batch dimension labels. This is necessary because the labels on the
        left- and right-hand sides of the equations need to agree -- since we've switched the core and batch on the
        left, we also have to do it on the right.
    - Finally, we can multiply (X * phi) and sum across the batch dimension

    This is all necessary because from the supply perspective, the 2nd dimension of X is the core dimension.
    """
    dims = list(dims)
    if sum_dim is None:
        sum_dim = dims[-1]
    sum_axis = dims.index(sum_dim)

    def _swapaxes(x, i, j):
        sub_template = Template("$x.subs({$i:$j})")
        return sub_template.safe_substitute(x=x, i=i, j=j)

    def _T(x, i, j, coords):
        named_dims = set(list(coords.keys()))

        # Use the first lowercase letter that hasn't been declared as a dim by the user as a temp
        temp_dim = sorted(list(set(string.ascii_lowercase) - named_dims))[0]
        return x + f".subs([({i}, {temp_dim}), ({j}, {i}), ({temp_dim}, {j})])"

    factors, factor_prices, factor_shares = map(
        unwrap_singleton_list, (factors, factor_prices, factor_shares)
    )

    if backend == "numba":
        batch_dim = dims.pop(sum_axis)
        core_dim = dims[0]
        n_core, n_batch = (len(coords[x]) for x in [core_dim, batch_dim])
        profit_output = output
        profit_price = output_price

        if transpose_output:
            profit_output = _swapaxes(output, batch_dim, core_dim)
            profit_price = _swapaxes(output_price, batch_dim, core_dim)

        zero_profit = f"{profit_output} = Sum({factor_prices} * {factors}, ({batch_dim}, 0, {n_batch - 1})) / {profit_price}"
        # zero_profit = f"{_swapaxes(output, batch_dim, core_dim)} = Sum({factor_prices} * {factors}, ({batch_dim}, 0, {n_batch - 1})) / {_swapaxes(output_price, batch_dim, core_dim)}"
        factor_demands = f"{factors} = {factor_shares} * {output}"

    elif backend == "pytensor":
        price_slice = "[:, None]" if expand_price_dim else ""
        zero_profit = f"{output} = ({factor_prices}{price_slice} * {factors}).sum(axis={sum_axis}).ravel() / ({output_price})"
        factor_demands = f"{factors} = {factor_shares} * {output}[None]"

    else:
        raise ValueError(f"backend must be one of 'numba' or 'pytensor', found {backend}")

    return zero_profit, factor_demands


def leontief(
    factors: str | list[str, ...],
    factor_prices: str | list[str, ...],
    output: str,
    output_price: str,
    factor_shares: str | list[str, ...],
    dims: str | list[str, ...],
    coords: dict[str, list[str, ...]],
    sum_dim: str | None = None,
    expand_price_dim: bool = True,
    transpose_output: bool = True,
    backend: Literal["numba", "pytensor"] = "numba",
) -> tuple[str, ...]:
    """
    Generate string equations representing a Leontief production process.

    A Leontief production process is defined as:

    .. math::
        Y = \\min_i \\left( \\frac{X_i}{\\alpha_i} \\right)

    Where :math:`\\alpha_i` is the share of factor :math:`i` in the production process. This generates the following
    factor demands:

    .. math::
        X_i = \\alpha_i Y

    Because of the non-linearity of the minimum operator, this function does not return the production function itself.
    Insead, we return the equivalent zero-profit condition:

    .. math::
        P_Y Y = \\sum_i P_i X_i

    Parameters
    ----------
    factors: list of str
        Input factors to the production process.

    factor_prices: list of str
        Production factor prices.

    output: str
        Name of the output of the production function

    output_price: str
        Name of the price of the output

    factor_shares: list of str
        Factor shares in the production function.

    dims: Sequence of str or str
        Sequence of dimensions indexing inputs and outputs into the Leontief production process. If a string or a list
        of length one is provided, the inputs and outputs will be assumed to be indexed by the provided dimension.

        If length == 2, the *second* dimension will be assumed to index the outputs, and the inputs will be reduced over
        the first dimension. This is the most common case in CGE modeling, when inputs to an intermediate consumption/
        value chain process are represented by an N x N matrix, with sectoral supply on the first dimension and sectoral
        demand on the second dimension.

    sum_dim: str, optional
        Dimension to reduce via summation. Default is the last dimension -- dims[-1]. Ignored if len(dims) == 1

    coords: dict of str: list of str
        Dictionary of coordinates for the model, mapping dimension names to lists of labels.

    backend: str, one of 'numba' or 'pytensor'
        Backend that will parse the equations. Only relevant for aggregation functions like Sum and Prod. Elementwise
        operations are interoperable between backends.

    Returns
    -------
    (zero_profit_constraint, factor_demands): tuple of strings
        Equations representing the Leontief production function and resulting factor demands
    """

    _check_pairwise_lengths_match(
        ["factors", "factor_prices", "factor_shares"],
        [factors, factor_prices, factor_shares],
    )

    factors = at_least_list(factors)
    if isinstance(dims, str) or len(dims) == 1:
        if len(factors) == 1:
            raise ValueError(
                f"Leontief production function expects at least two factors when len(dims) == 1, found {len(factors)}"
            )
        return _1d_leontief(
            factors,
            factor_prices,
            output,
            output_price,
            factor_shares,
            dims,
            coords,
            backend,
        )
    else:
        if len(factors) != 1:
            raise ValueError(
                f"Leontief production function expects exactly one factor when len(dims) == 2, found {len(factors)}"
            )
        return _2d_leontief(
            factors=factors,
            factor_shares=factor_shares,
            output=output,
            output_price=output_price,
            factor_prices=factor_prices,
            dims=dims,
            coords=coords,
            sum_dim=sum_dim,
            expand_price_dim=expand_price_dim,
            transpose_output=transpose_output,
            backend=backend,
        )



import functools as ft

import numpy as np
import pytensor
import pytensor.tensor as pt

from cge_modeling.pytensorf.compile import flat_tensor_to_ragged_list
from cge_modeling.tools.pytensor_tools import flatten_equations


def eval_func_maybe_exog(X, exog, f, has_exog):
    if has_exog:
        out = f(*X, *exog)
    else:
        out = f(*X)

    return out


def _newton_step(flat_X, exog, F, J, step_size, has_exog, shapes):
    X = flat_tensor_to_ragged_list(flat_X, shapes)
    F_X = eval_func_maybe_exog(X, exog, F, has_exog)
    J_X = eval_func_maybe_exog(X, exog, J, has_exog)

    flat_F_X = flatten_equations(F_X)
    flat_new_X = flat_X - step_size * pt.linalg.solve(J_X, flat_F_X)

    new_X = flat_tensor_to_ragged_list(flat_new_X, [x.type.shape for x in X])
    flat_F_new_X = eval_func_maybe_exog(new_X, exog, F, has_exog)

    return flat_X, flat_new_X, flat_F_X, flat_F_new_X


def no_op(X):
    return X, X, X, X


def compute_norms(X, new_X, F_X, F_new_X):
    norm_X = pt.linalg.norm(X, ord=1)
    norm_new_X = pt.linalg.norm(new_X, ord=1)
    norm_root = pt.linalg.norm(F_X, ord=1)
    norm_root_new = pt.linalg.norm(F_new_X, ord=1)
    norm_step = pt.linalg.norm(new_X - X, ord=1)

    return norm_X, norm_new_X, norm_root, norm_root_new, norm_step


def _check_convergence(norm_step, norm_root, tol):
    #     new_converged = pt.or_(pt.lt(norm_step, tol), pt.lt(norm_root, tol))
    new_converged = pt.lt(norm_root, tol)
    return new_converged


def check_convergence(norm_step, norm_root, converged, tol):
    return pytensor.ifelse(converged, np.array(True), _check_convergence(norm_step, norm_root, tol))


def check_stepsize(norm_root, norm_root_new, step_size, initial_step_size):
    is_decreasing = pt.lt(norm_root_new, norm_root)

    return pytensor.ifelse(
        is_decreasing,
        (is_decreasing, initial_step_size),
        (is_decreasing, step_size * 0.5),
    )


def backtrack_if_not_decreasing(is_decreasing, X, new_X):
    return pytensor.ifelse(is_decreasing, new_X, X)


def scan_body(*args, F, J, initial_step_size, tol, has_exog, n_endog, n_exog):
    X = args[:n_endog]
    converged, step_size, n_steps = args[n_endog : n_endog + 3]
    exog = args[-n_exog:]

    shapes = [x.type.shape for x in X]
    flat_X = flatten_equations(X)

    out = pytensor.ifelse(
        converged,
        no_op(flat_X),
        _newton_step(flat_X, exog, F, J, step_size, has_exog, shapes),
    )

    flat_X, flat_new_X, flat_F_X, flat_F_new_X = (out[i] for i in range(4))
    norm_X, norm_new_X, norm_root, norm_root_new, norm_step = compute_norms(
        flat_X, flat_new_X, flat_F_X, flat_F_new_X
    )
    is_converged = check_convergence(norm_step, norm_root, converged, tol)

    is_decreasing, new_step_size = check_stepsize(
        norm_root, norm_root_new, step_size, initial_step_size
    )

    flat_return_X = backtrack_if_not_decreasing(is_decreasing, flat_X, flat_new_X)
    return_X = flat_tensor_to_ragged_list(flat_return_X, shapes)
    new_n_steps = n_steps + (1 - is_converged)

    return [*return_X, is_converged, new_step_size, new_n_steps]


def _process_root_data(data: dict[str, np.ndarray] | None) -> list[pt.TensorLike]:
    if data is None:
        return [pt.as_tensor_variable(np.nan, name="dummy_exog", shape=())]
    else:
        out = []
        for name in data.keys():
            x = np.array(data[name], dtype=pytensor.config.floatX)
            out.append(pt.as_tensor_variable(x, name=name, shape=x.shape))
        return out


def root(
    f: pt.Op,
    f_jac: pt.Op,
    initial_data: dict[str, np.ndarray | float],
    parameters: dict[str, np.ndarray | float] | None = None,
    step_size: int = 1,
    max_iter: int = 100,
    tol: float = 1e-8,
) -> tuple[list[pt.TensorVariable], pt.TensorVariable, pt.TensorVariable, pt.TensorVariable]:
    """
    Find the root of a system of equations using Newton's method with backtracking.

    Parameters
    ----------
    f: pytensor Op
        A pytensor Op, typically created by CGEModel.compile_equations_to_pytensor, that takes a ragged list of
        variables and parameters as input and returns a vector of residuls as output.
    f_jac: pytensor Op
        A pytensor Op, typically created by CGEModel.compile_equations_to_pytensor, that takes a ragged list of
        variables and parameters as input and returns the inverse of the Jacobian of the system of equations as output.
    initial_data: dict[str, np.ndarray]
        A dictionary of initial values for the variables in the system of equations. The keys are the names of the
        variables and the values are the initial values.
    parameters: Optional, dict[str, np.ndarray]
        A dictionary of exogenous parameters for the system of equations. The keys are the names of the parameters and
        the values are the parameter values.
    step_size: int
        The initial step size to use in Newton's method
    max_iter: int
        The maximum number of iterations to run Newton's method
    tol: float
        The tolerance for convergence. The algorithm will stop if the norm of the residuals is less than the tolerance
        or if the norm of the step size is less than the tolerance.

    Returns
    -------
    root_histories: list[pt.TensorVariable]
        A list of the values of the variables at each iteration of the algorithm.

    converged: pt.TensorVariable
        A boolean tensor indicating whether the algorithm converged at each iteration.

    step_size: pt.TensorVariable
        The step size used at each iteration. Useful to diagnose whether the algorithm is converging.

    n_steps: pt.TensorVariable
        The number of steps taken at each iteration. Useful to diagnose whether the algorithm is converging.
    """

    init_step_size = np.float64(step_size)
    converged = np.array(False)
    n_steps = 0
    has_exog = parameters is not None

    x0 = _process_root_data(initial_data)
    exog = _process_root_data(parameters)

    n_endog = len(x0)
    n_exog = len(exog)

    root_func = ft.partial(
        scan_body,
        F=f,
        J=f_jac,
        initial_step_size=init_step_size,
        tol=tol,
        has_exog=has_exog,
        n_endog=n_endog,
        n_exog=n_exog,
    )

    outputs, updates = pytensor.scan(
        root_func,
        outputs_info=[*x0, converged, init_step_size, n_steps],
        non_sequences=exog,
        n_steps=max_iter,
        strict=True,
    )

    *root_histories, converged, step_size, n_steps = outputs

    return root_histories, converged, step_size, n_steps



from IPython import display


def print_latex(eqs):
    for eq in eqs:
        display(eq)


def union_many(x, *args):
    for y in args:
        if not isinstance(y, list | tuple | set):
            y = (y,)
        x = x.union(set(y))
    return x


def brackets_to_snake_case(name):
    """
    Rename a variable x[i, j] to x_i_j
    """
    name = name.replace("]", "")
    name = name.replace("[", "_")
    name = name.replace(",", "_")
    name = name.replace(" ", "")
    return name



import functools as ft
import re
import sys

from collections.abc import Callable, Sequence
from itertools import product
from typing import Any, cast

import numpy as np

from fastprogress.fastprogress import ProgressBar, progress_bar

from cge_modeling.base.primitives import (
    Equation,
    Parameter,
    Variable,
    _pretty_print_dim_flags,
)

CGETypes = Variable | Parameter | Equation


def unpack_equation_strings(
    equations: list[Equation, ...], coords: dict[str, list[str | int, ...]]
) -> list[str, ...]:
    """
    Helper function to unpack equations names with <dim:i> tags to a long list of equations.

    Parameters
    ----------
    equations: Equation
        List of equation objects from a CGE model
    coords: dict
        Dictionary mapping dimension names to  lists of labels

    Returns
    -------
    unpacked_names: list[str, ...]
        List of unpacked equation names
    """

    unpacked_names = []
    dims = list(coords.keys())
    dims_pattern = "|".join([re.escape(x) for x in dims])
    pattern = f"<dim:({dims_pattern})>"

    for eq in equations:
        name = eq.name
        named_dims = re.findall(pattern, name)

        if len(named_dims) == 0:
            unpacked_names.append(name)
            continue
        labels = [coords[dim] for dim in named_dims]
        label_prod = product(*labels)
        for labels in label_prod:
            sub_dict = dict(zip(named_dims, labels))
            new_name = _pretty_print_dim_flags(name, named_dims, sub_dict)
            unpacked_names.append(new_name)

    return unpacked_names


def _validate_input(obj: Any, cls: CGETypes):
    """
    Validate that an input is an instance of a class.

    Parameters
    ----------
    obj: Any
        An object to validate
    cls: Union[Variable, Parameter]
        The class to validate against
    """

    if not isinstance(obj, cast(type, cls)):
        raise ValueError(f"Expected instance of type {cls.__name__}, found {type(obj).__name__}")


def ensure_input_is_sequence(x: Any) -> Sequence[Any]:
    """
    Convert non-sequence inputs into a singleton list.

    Parameters
    ----------
    x: Any
        An anonymous input

    Returns
    -------
    x: list[Any]
        A list containing the input, or the input itself if it is already a sequence
    """
    if not isinstance(x, list | tuple):
        x = [x]
    return x


def _expand_var_by_index(
    obj: Variable | Parameter, coords: dict[str, list[str | int, ...]]
) -> list[Variable | Parameter]:
    """
    Create a set of CGE Model objects from a single object using the cartesian product of the object's dimensions

    Parameters
    ----------
    obj: Union[Variable, Parameter]
        A CGE Model object to be expanded
    coords: dict[str, list[str]]
        A dictionary of coordinates known by the model. Used to determine the number and labels of the new objects.

    Returns
    -------
    output_set: list[Union[Variable, Parameter]]
        A list of new CGE Model objects, one for each combination of dimensions in the input object.

    Notes
    -----
    The new objects will have the same name, description, and type as the input object, but the abstract dimension names
    will be replaced with actual labels from the coordinates dictionary.

    Examples
    --------
    .. code-block:: python

        from cge_modeling.base.primitives import Variable
        from cge_modeling.base.utilities import _expand_var_by_index

        x = Variable(name='x', dims='i', description='Quantity associated with sector <dim:i> ')
        coords = {'i': ['Agriculture', 'Industrial', 'Service']}
        expanded_x = _expand_var_by_index(x, coords)
        for var in expanded_x:
            print(var._full_latex_name)
        # Out: x_{i=\text{Agriculture}}
        #      x_{i=\text{Industrial}}
        #      x_{i=\text{Service}}
    """
    if not isinstance(obj, Variable | Parameter):
        raise ValueError(f"Expected a model object for argument obj, got {type(obj)}")

    dims = list(coords.keys())
    missing_dims = set(dims) - set(obj.dims)
    if len(missing_dims) > 0:
        raise ValueError(
            f'Found indices {", ".join(missing_dims)} on the coords that are not among the variable '
            f'indices: {", ".join(obj.dims)}'
        )

    cls = Variable if isinstance(obj, Variable) else Parameter
    output_set = [obj]

    for dim in dims:
        new_out = []
        for obj in output_set:
            labels = coords.get(dim, []) if dim in obj.dims else []
            for label in labels:
                new_dim_vals = obj.dim_vals.copy()
                new_dim_vals.update({dim: label})

                # noinspection PyArgumentList
                new_out.append(
                    cls(
                        name=obj.name,
                        dims=obj.dims,
                        dim_vals=new_dim_vals,
                        description=obj.description,
                    )
                )
        output_set = new_out.copy()

    return output_set


def _replace_dim_marker_with_dim_name(s: str) -> str:
    """
    Strip the dimension tag <dim:> from a description string, returning only the dimension label itself.

    Parameters
    ----------
    s: str
        A string containing a dimension marker, e.g. "<dim:i>"

    Returns
    -------
    new_s: str
        The input string with the dimension marker removed, e.g. "i"

    Notes
    -----
    The dimension markers used in description strings are ugly, but they are an easy way to facilitate swapping in
    arbitrary dimension labels when rendering equations to latex. Nevertheless, users will not want to see them when
    inspecting the "compact" form of the model. This function removes them.

    Examples
    --------
    .. code-block:: python

        from cge_modeling.base.utilities import _replace_dim_marker_with_dim_name

        s = 'Sector <dim:i> demand for good <dim:j>'
        _replace_dim_marker_with_dim_name(s)
        # Out: 'Sector i demand for good j'
    """
    s = re.sub("(<dim:(.+?)>)", r"\g<2>", s)
    return s


def infer_object_shape_from_coords(
    obj: Variable | Parameter, coords: dict[str, list[str | int, ...]]
) -> tuple[int]:
    """
    Infer the shape of a CGE Model object from a provided coordinate dictionary.

    Parameters
    ----------
    obj: Union[Variable, Parameter]
        A CGE Model object
    coords: dict[str, list[str]]
        A dictionary of coordinates, mapping dimension names to lists of labels associated with that dimension.

    Returns
    -------
    shape: tuple[int]
        A tuple of integers representing the shape of the object.

    Notes
    -----
    CGE Model objects do not themselves carry shape information, only dimension information. This is because the shape
    information is not strictly necessary at model declaration time, and indeed might not even be known by the economist
    when designing the model.

    What will determine the shape information is the coordinate dictionary, which is known by the CGEModel object but
    not the individual Variable and Parameter instances it contains. This function allows us to infer the shape of an
    object from the coordinate dictionary.

    Examples
    --------
    .. code-block:: python
        from cge_modeling.base.primitives import Variable
        from cge_modeling.base.utilities import infer_object_shape_from_coords

        x = Variable(name='x', dims='i', description='Quantity associated with sector <dim:i> ')
        coords = {'i': ['Agriculture', 'Industrial', 'Service']}
        infer_object_shape_from_coords(x, coords)
        # Out: (3,)

        z = Variable(name='phi_X', dims='i, j', description='Quantity of <dim:j> goods in sector <dim:i> value chain')
        coords = {'i': ['Agriculture', 'Industrial', 'Service'], 'j': ['Agriculture', 'Industrial', 'Service']}
        infer_object_shape_from_coords(z, coords)
        # Out: (3, 3)
    """
    dims = obj.dims
    shape = tuple(len(coords[dim]) for dim in dims)
    return shape


def make_flat_array_return_mask(
    x: np.ndarray,
    all_objects: list[Variable | Parameter],
    omit_object_names: list[str],
    coords: dict[str, list[str | int, ...]],
) -> np.ndarray:
    mask = np.full(x.shape[0], True)
    cursor = 0
    for obj in all_objects:
        shape = infer_object_shape_from_coords(obj, coords)
        s = int(np.prod(shape))

        if obj.name in omit_object_names:
            mask[cursor : cursor + s] = False
        cursor += s

    return mask.astype("bool")


def flat_array_to_variable_dict(
    x: np.ndarray,
    objects: list[Variable | Parameter],
    coords: dict[str, list[str | int, ...]],
) -> dict[str, np.ndarray]:
    """
    Convert a flat array to a dictionary of variables and parameters.

    Parameters
    ----------
    x: np.ndarray
        Flat array containing all variable and parameters in the model. The ordering is assumed to match the ordering
        given by list passed to the objects argument, with multidimensional objects flattened in row-major order.
    objects: list[Union[Variable, Parameter]]
        List of variables and parameters
    coords: dict[str, list[str]]
        Dictionary of coordinates mapping dimension names to lists of labels associated with that dimension.

    Returns
    -------
    d: dict[str, np.ndarray]
        Dictionary mapping variable and parameter names to numpy arrays containing the values of those variables and
        parameters.

    Notes
    -----
    This function is the inverse of variable_dict_to_flat_array.

    The goal of this pair of functions is to hide  flattening and concatenation operations from the user. The user
    should only have to reason about input and outputs in terms of each individual object, without ever having to
    worry about unifying them into a single long vector or large matrix.

    The ordering of the variables and parameters in the output vector is determined by the order of the variables and
    parameters in the variable_list and parameter_list arguments. Variables are ordered first, followed by parameters.
    Multidimensional objects are flattened in row-major order.

    # TODO: This function cannot currently handle batch dimensions, but it should be able to.

    Examples
    --------
    .. code-block:: python
        from cge_modeling.base.primitives import Variable
        from cge_modeling.base.utilities import flat_array_to_variable_dict

        x = Variable(name='x', dims='i', description='Output of the <dim:i> sector')
        coords = {'i': ['Agriculture', 'Industrial', 'Service']}
        flat_array_to_variable_dict(np.array([1, 2, 3]), [x], coords)
        # Out: {'x': array([1, 2, 3])}

        z = Variable(name='phi_X', dims='i, j', description='Quantity of <dim:j> goods in sector <dim:i> value chain')
        coords = {'i': ['Agriculture', 'Industrial', 'Service'], 'j': ['Agriculture', 'Industrial', 'Service']}
        flat_array_to_variable_dict(np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]), [z], coords)
        # Out: {'phi_X': array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])}
    """
    d = {}
    cursor = 0
    for obj in objects:
        shape = infer_object_shape_from_coords(obj, coords)
        s = int(np.prod(shape))
        value = x[cursor : cursor + s].reshape(shape)
        d[obj.name] = value
        cursor += s

    return d


def variable_dict_to_flat_array(
    d: dict[str, np.ndarray],
    variable_list: [list[Variable]],
    parameter_list: list[Parameter],
) -> tuple[np.ndarray, np.ndarray]:
    """
    Convert a dictionary of variables and parameters to a single long vector.

    Parameters
    ----------
    d: dict[str, np.ndarray]
        Data dictionary mapping variable and parameter names to numpy arrays. Numpy arrays can be any dimension, and are
        assumed to have dims corresponding to their object's dims.

    variable_list: list[Variable]
        List of variables in the model
    parameter_list: list[Parameter]
        List of parameters in the model

    Returns
    -------
    variables: np.ndarray
        A single long vector containing all variables in the model

    parameters: np.ndarray
        A single long vector containing all parameters in the model

    Notes
    -----
    This function is the inverse of flat_array_to_variable_dict.

    The goal of this pair of functions is to hide  flattening and concatenation operations from the user. The user
    should only have to reason about input and outputs in terms of each individual object, without ever having to
    worry about unifying them into a single long vector or large matrix.

    The ordering of the variables and parameters in the output vector is determined by the order of the variables and
    parameters in the variable_list and parameter_list arguments. Variables are ordered first, followed by parameters.
    Multidimensional objects are flattened in row-major order.

    # TODO: This function cannot currently handle batch dimensions, but it should be able to.
    """

    variables = np.concatenate([np.atleast_1d(d[var.name]).ravel() for var in variable_list])
    parameters = np.concatenate([np.atleast_1d(d[var.name]).ravel() for var in parameter_list])

    return variables, parameters


def wrap_fixed_values(
    f: Callable,
    fixed_values: dict[str, float | int | np.ndarray],
    variables: list[Variable],
    coords: dict[str, list[str | int, ...]],
) -> Callable:
    """
    Wrap a CGE function to require only a subset of its inputs. Useful when optimizing under a constraint that
    certain variables are fixed at initial values.

    Parameters
    ----------
    f: Callable
        A compiled CGE function with signature f(**variables, **parameters)

    fixed_values: dict[str, Union[float, int, np.ndarray]]
        Dictionary mapping variable names to numpy arrays containing the fixed values of those variables. Keys should
        be a subset of the variable names accepted by the function f.

    variables: list[Variable]
        List of model variables

    coords: dict[str, list[str, ...]]
        Dictionary of coordinates mapping dimension names to lists of labels associated with that dimension.

    Returns
    -------
    inner: Callable
        A wrapped version of the input function that accepts only the variables not in fixed_values as keyword arguments.
    """
    var_names = [var.name for var in variables]
    if any(var not in var_names for var in fixed_values.keys()):
        raise ValueError(
            "User asked to fix the following variables, but these variables are not in the model: "
            f"{set(fixed_values.keys()) - set(var_names)}"
        )
    fixed_vars = [var.name for var in variables if var.name in fixed_values.keys()]

    @ft.wraps(f)
    def inner(**data):
        if any(var in fixed_vars for var in data.keys()):
            raise ValueError(
                f"Values for the following variables were passed to a function that expected them to be "
                f"fixed: {set(fixed_vars) & set(data.keys())}"
            )
        data.update(fixed_values)
        res = f(**data)
        if isinstance(res, float | int) or res.ndim == 0:
            return res

        return_mask = make_flat_array_return_mask(res, variables, fixed_vars, coords)
        if res.ndim == 1:
            return res[return_mask]
        return res[return_mask, :][:, return_mask]

    return inner


def wrap_pytensor_func_for_scipy(
    f: Callable,
    variable_list: list[Variable],
    parameter_list: list[Parameter],
    coords: dict[str, list[str | int, ...]],
) -> Callable:
    """
    Wrap a PyTensor function for use with scipy.optimize.root or scipy.optimize.minimize.

    Parameters
    ----------
    f: Callable
        A compiled PyTensor function with an input signature f(**data), where data is a dictionary mapping variable
        and parameter names to numpy arrays containing the values of those variables and parameters.

    variable_list: list[Variable]
        List of model variables
    parameter_list: list[Parameter]
        List of model parameters
    coords: dict[str, list[str, ...]]
        Dictionary of coordinates mapping dimension names to lists of labels associated with that dimension.

    Returns
    -------
    inner_f: Callable
        A wrapped version of the input function that accepts a single long vector of variables as input, and a single
        long vector of parameters as second input. The wrapped function will unpack these vectors into a dictionary of
        variables and parameters, and then unpack that dictionary into keyword arguments to the original function.
    """

    @ft.wraps(f)
    def inner_f(x0, theta):
        # Scipy will pass x0 as a single long vector, and theta separately (but also as a single long vector).
        inputs = np.r_[x0, theta]
        data = flat_array_to_variable_dict(inputs, variable_list + parameter_list, coords)
        return f(**data)

    return inner_f


def flat_mask_from_param_names(param_dict, names):
    keys = list(param_dict.keys())
    count_dict = {k: np.prod(np.atleast_1d(v).shape) for k, v in param_dict.items()}
    counts = np.concatenate([np.atleast_1d(count_dict[var]).ravel() for var in keys])
    cumcounts = counts.cumsum()

    slice_dict = count_dict.copy()

    for i, k in enumerate(keys):
        slice_dict[k] = slice(cumcounts[i] - counts[i], cumcounts[i])

    flat_data = np.concatenate([np.atleast_1d(param_dict[var]).ravel() for var in keys])
    mask = np.full(flat_data.shape, False, dtype="bool")

    if isinstance(names, str):
        names = [names]

    for name in names:
        mask[slice_dict[name]] = True

    return mask


def _optimzer_early_stopping_wrapper(f_optim):
    objective: CostFuncWrapper = f_optim.args[0]
    progressbar = objective.progressbar

    res = f_optim()
    res.args = objective.args
    total_iters = objective.n_eval

    if progressbar:
        out = objective.step(res.x)

        if not isinstance(out, tuple):
            out = (out,)
        if objective.use_hess:
            out += (objective.f_hess(res.x),)

        objective.update_progress_desc(*out)
        objective.progress.total = total_iters
        objective.progress.update(total_iters)

        print(file=sys.stdout)

    return res


class CostFuncWrapper:
    def __init__(
        self,
        f: Callable,
        args: tuple | None,
        f_jac: Callable | None = None,
        f_hess: Callable | None = None,
        maxeval: int = 5000,
        tol: float = 1e-6,
        progressbar: bool = True,
        update_every: int = 10,
        backend="numba",
    ):
        kwargs = dict(theta=args) if backend == "pytensor" else dict(endog_inputs=args)

        self.n_eval = 0
        self.maxeval = maxeval
        self.tol = tol
        self.f = f if args is None else ft.partial(f, **kwargs)
        self.args = args

        self.use_jac = False
        self.use_hess = False
        self.update_every = update_every
        self.interrupted = False
        self.desc = "f = {:,.5g}"

        if f_jac is not None:
            self.desc += ", ||grad|| = {:,.5g}"
            self.use_jac = True
            self.f_jac = f_jac if args is None else ft.partial(f_jac, **kwargs)

        if f_hess is not None:
            self.desc += ", ||hess|| = {:,.5g}"
            self.use_hess = True
            self.f_hess = f_hess if args is None else ft.partial(f_hess, **kwargs)

        self.previous_x = None
        self.progressbar = progressbar
        self.progress = None

        if progressbar:
            self.progress = progress_bar(range(maxeval), total=maxeval, display=progressbar)
            self.progress.update(0)

    def step(self, x):
        grad = None
        hess = None
        value = self.f(x)

        if self.use_jac:
            grad = self.f_jac(x)
            if self.use_hess:
                hess = self.f_hess(x)
            if np.all(np.isfinite(x)):
                self.previous_x = x
        else:
            self.previous_x = x

        if self.n_eval % self.update_every == 0:
            self.update_progress_desc(value, grad, hess)

        if self.n_eval > self.maxeval:
            self.update_progress_desc(value, grad, hess)
            self.interrupted = True
            if self.use_jac:
                return value, grad
            return np.array(value)

        self.n_eval += 1
        if self.progressbar:
            assert isinstance(self.progress, ProgressBar)
            self.progress.update_bar(self.n_eval)

        if self.use_jac:
            if self.use_hess:
                return value, grad  # , hess
            else:
                return value, grad
        else:
            return np.array(value)

    def __call__(self, x):
        try:
            new_x = self.step(x)
            value = np.array(new_x[0]) if isinstance(new_x, tuple) else np.array(new_x)
            self.interrupted = (self.n_eval > self.maxeval) or (value < self.tol)
            return new_x

        except (KeyboardInterrupt, StopIteration):
            self.interrupted = True
            # Call step again so there's a chance for the callback to trigger before we just raise StopIteration
            # In optimize.minimize, this lets us break out and keep the optimizer state
            return self.step(x)

    def callback(self, x):
        if self.interrupted:
            raise StopIteration

    def update_progress_desc(
        self,
        value: float,
        grad: np.ndarray | None = None,
        hess: np.ndarray | None = None,
    ) -> None:
        if not self.progressbar:
            return

        value = np.array(value)
        if value.shape != ():
            value = (value**2).sum() / min(1, int(np.prod(value.shape)))

        if grad is None:
            self.progress.comment = self.desc.format(value)
        else:
            if hess is None:
                norm_grad = np.linalg.norm(grad)
                self.progress.comment = self.desc.format(value, norm_grad)
            else:
                norm_grad = np.linalg.norm(grad)
                norm_hess = np.linalg.norm(hess)
                self.progress.comment = self.desc.format(value, norm_grad, norm_hess)



import numpy as np
import pytensor
import pytensor.tensor as pt

from cge_modeling.pytensorf.compile import euler_approximation
from cge_modeling.tools.pytensor_tools import make_jacobian


def test_euler_approximation_1d():
    x = pt.dscalar("x")
    y = pt.dscalar("y")
    eq = pt.atleast_1d(y - pt.cos(x) + 1)
    n_steps = pt.iscalar("n_steps")

    A = make_jacobian(eq, [y])
    B = make_jacobian(eq, [x])

    x0_final, result = euler_approximation(A, B, variables=[y], parameters=[x], n_steps=n_steps)
    f = pytensor.function([x, y, x0_final, n_steps], result)
    y_values, x_values = f(0, 0, np.array([10.0]), 10_000)
    true = -np.cos(np.array([10])) + 1
    np.testing.assert_allclose(-true, y_values[-1], atol=1e-3)


def test_euler_approximation_2d():
    variables = v1, v2 = [pt.dscalar(name) for name in ["v1", "v2"]]
    parameters = v3 = pt.dscalar("v3")
    n_steps = pt.iscalar("n_steps")
    inputs = [*variables, parameters]

    equations = pt.stack([v1**2 * v3 - 1, v1 + v2 - 2])
    A = make_jacobian(equations, variables)
    B = make_jacobian(equations, [parameters])

    theta_final, result = euler_approximation(
        A, B, variables=variables, parameters=[parameters], n_steps=n_steps
    )
    f = pytensor.function([*inputs, theta_final, n_steps], result)

    def f_analytic(v3):
        v1 = 1 / np.sqrt(v3)
        v2 = 2 - v1
        return np.array([v1, v2])

    initial_point = [1.0, 1.0]
    v3_initial = 1.0
    v3_final = 2.0

    analytic_solution = np.array(f_analytic(v3_final))
    *x, theta = f(*initial_point, v3_initial, np.array([v3_final]), 10_000)
    np.testing.assert_allclose(x[0][-1], analytic_solution[0], atol=1e-3)
    np.testing.assert_allclose(x[1][-1], analytic_solution[1], atol=1e-3)


def test_euler_approximation_step():
    pass



from typing import Any, Literal, cast

import numpy as np
import pytensor
import pytensor.tensor as pt
import sympy as sp

from pytensor.compile.sharedvalue import SharedVariable
from pytensor.graph import rewrite_graph
from pytensor.graph.basic import graph_inputs
from sympytensor import as_tensor

from cge_modeling.base.primitives import Parameter, Variable
from cge_modeling.tools.sympy_tools import sparse_jacobian


def object_to_pytensor(obj: Parameter | Variable, coords: dict[str, list[str, ...]]):
    """
    Convert a CGE model object to a PyTensor object.

    Parameters
    ----------
    obj: Union[Parameter, Variable]
        The CGE model object to convert
    coords: dict[str, Sequence]
        A dictionary of coordinate names and values known by the CGE model. Used to determine the shapes and dimensions
        of the returned PyTensor object

    Returns
    -------
    pytensor_obj: pytensor.tensor.TensorVariable
        The PyTensor object representing the CGE model object

    Notes
    -----
    Distinction between Parameters and Variables is lost in the conversion to symbolic tensors. It is up to the user to
    maintain this distinction.

    Examples
    --------
    .. code-block:: python

        from cge_modeling.base.primitives import Parameter
        from cge_modeling.pytensorf.compile import object_to_pytensor

        K_d = Parameter(name='K_d', dims='i', description='Capital demand')
        coords = {'i': [0, 1, 2]}
        pt_K_d = object_to_pytensor(K_d, coords)
        pt_K_d.eval({pt_K_d: [1, 2, 3]})
        # Out: array([1., 2., 3.])
    """

    name = obj.name
    dims = obj.dims

    shape = tuple(list(map(lambda dim: len(coords[dim]), dims)))

    return pt.tensor(name, shape=shape)


def make_printer_cache(variables: list[pt.TensorLike], parameters: list[pt.TensorLike]) -> dict:
    """
    Create a cache of PyTensor functions for printing sympy equations to pytensor.

    Parameters
    ----------
    cge_model: CGEModel
        CGEModel object
    variables: list of TensorVariable
        List of cge_model variables, converted to symbolic pytensor tensors
    parameters: list of TensorVariable
        List of cge_model parameters, converted to symbolic pytensor tensors

    Returns
    -------
    cache: dict
        A cache of variables used to print equations to pytensor
    """

    def make_key(var) -> tuple:
        return var.name, sp.Symbol, (), "floatX", ()

    cache = {make_key(var): var for var in variables + parameters}
    return cache


def flatten_equations(eqs: list[pt.TensorLike]) -> pt.TensorLike:
    """
    Flatten a list of pytensor vectors, each representing equations in a system of equations,  into a single vector.

    Parameters
    ----------
    eqs: list[pytensor.tensor.TensorVariable]
        A list of pytensor vectors, each representing equations in a system of equations

    Returns
    -------
    flat_equations: pytensor.tensor.TensorVariable
        A single pytensor vector representing the equations in the system of equations

    Notes
    -----
    In the context of CGE modeling, each equation in the model can be thought of as a vector (or matrix) of equations,
    with each dimension corresponding to a single variable index. This perspective is useful for efficient vectorized
    computation of each "block" of the economic model. However, when solving for the equilibrium, we need to flatten
    these multidimensional expressions into a single vector to compute the Jacobian of the system. This function
    performs that flattening operation.
    """
    expr_len = 0
    for eq in eqs:
        if isinstance(eq, pt.TensorVariable):
            expr_len += int(np.prod(eq.type.shape))
        else:
            expr_len += int(np.prod(np.atleast_1d(eq).size))

    # expr_len = int(np.sum([np.prod(eq.type.shape) for eq in eqs]))
    flat_expr = pt.concatenate(
        [pt.atleast_1d(pt.cast(eq, pytensor.config.floatX)).ravel() for eq in eqs],
        axis=-1,
    )
    flat_expr = pt.specify_shape(flat_expr, shape=(expr_len,))
    return flat_expr


def unpacked_graph_to_packed_graph(
    unpacked_graph: pt.TensorVariable | list[pt.TensorVariable],
    cge_model,
    cache,
    unpacked_cache,
) -> pt.TensorVariable | list[pt.TensorVariable]:
    pt_vars = list(cache.values())
    cache_var_names = list(x[0] for x in cache.keys())
    unpacked_to_indexed_dict = {}

    for info, var in unpacked_cache.items():
        name, *_ = info  # info is a tuple of (name, sympy_class, broadcastable, dtype, shape)
        parent_obj = cge_model.get_unpacked_parent_object(name)
        parent_pt = pt_vars[cache_var_names.index(parent_obj.base_name)]

        # Map named indices to integer indices
        dims = parent_obj.dims
        dim_idx = [cge_model.coords[dim].index(parent_obj.dim_vals[dim]) for dim in dims]

        if len(dim_idx) == 0:
            unpacked_to_indexed_dict[var] = parent_pt
            continue

        # Index the parent object with the integer indices
        indexed_parent = parent_pt[tuple(dim_idx)]
        name = f"{parent_pt.name}"
        name += f'[{", ".join([str(x) for x in dim_idx])}]' if len(dim_idx) > 0 else ""

        indexed_parent.name = name
        unpacked_to_indexed_dict[var] = indexed_parent

    # Replace the unpacked variables with the indexed variables
    packed_graph = pytensor.clone_replace(unpacked_graph, unpacked_to_indexed_dict)

    return packed_graph


def make_jacobian_from_sympy(
    cge_model,
    cache,
    unpacked_cache,
    wrt: Literal["variables", "parameters"] = "variables",
    sparse=False,
) -> pt.TensorVariable:
    equations = cge_model.unpacked_equation_symbols
    variables = cge_model.unpacked_variable_symbols
    parameters = cge_model.unpacked_parameter_symbols

    wrt_symbols = variables if wrt == "variables" else parameters
    if sparse:
        jac = sparse_jacobian(equations, wrt_symbols)
    else:
        jac = sp.Matrix(equations).jacobian(wrt_symbols)

    if wrt == "variables":
        cge_model.symbolic_solution_matrices["Jacobian"] = jac
    else:
        cge_model.symbolic_solution_matrices["B_matrix"] = jac

    unpacked_jac = as_tensor(jac, cache=unpacked_cache)
    packed_jac = unpacked_graph_to_packed_graph(unpacked_jac, cge_model, cache, unpacked_cache)

    return packed_jac


def make_jacobian(
    system: pt.TensorLike,
    x: list[pt.TensorLike],
    return_jvp: bool = False,
    p: None | pt.TensorVariable = None,
) -> pt.TensorVariable | tuple[pt.TensorVariable, pt.TensorVariable]:
    """
    Make a Jacobian matrix from a system of equations and a list of variables.

    Parameters
    ----------
    system: pytensor.tensor.TensorVariable
        A vector of equations
    x: list[pytensor.tensor.TensorVariable]
        A list of variables
    return_jvp: bool
        If true, compute only a jacobian-vector product with respect to an arbitray vectory p
    p: TensorVariable, optional
        Symbolic variable used in the computation of the JVP. Ignored if return_jvp is False. If None, a variable will
        be created.

    Returns
    -------
    jac: pytensor.tensor.TensorVariable
        The Jacobian matrix of the system of equations with respect to the variables
    p: pytensor.tensor.TensorVariable
        The symbolic variable used in the computation of the JVP. Only returned if return_jvp is True.

    Notes
    -----
    The Jacobian of the system of equations is the matrix of partial derivatives of each equation with respect to each
    variable. The rows of the matrix correspond to the equations in the system, while the columns correspond to the
    variables. This function computes the Jacobian matrix from a vector of equations and a list of variables.
    """
    n_eq = system.type.shape[0]
    n_vars = int(np.sum([np.prod(var.type.shape) for var in x]))

    rewrite_pregrad(system)

    if return_jvp:
        if p is None:
            p = [var.clone(name=f"{var.name}_point") for var in x]
        jvp = pytensor.gradient.Rop(system, x, p)
        return jvp, p

    column_list = pytensor.gradient.jacobian(system, x)

    jac = pt.concatenate([pt.atleast_2d(x).reshape((n_eq, -1)) for x in column_list], axis=-1)
    jac = pt.specify_shape(jac, shape=(n_eq, n_vars))

    return jac


def at_least_list(x: Any) -> list[Any]:
    """
    Wrap non-list objects in a list

    Parameters
    ----------
    x: Any
        An object to wrap in a list if it is not already a list

    Returns
    -------
    x: list
        The input object wrapped in a list if it was not already a list
    """
    if isinstance(x, list):
        return x
    else:
        return cast(list, [x])


def clone_and_rename(x: pt.TensorVariable, suffix: str = "_next") -> pt.TensorVariable:
    """
    Clone a pytensor tensor and rename it by appending a suffix to its name.

    Cloning variables is useful for creating intermediate "dummy" variables that can be replaced by graph rewrites,
    mostly commonly in vectorization.

    Parameters
    ----------
    x: pytensor.tensor.TensorVariable
        The pytensor variable to clone
    suffix: str
        The suffix to append to the cloned variable's name

    Returns
    -------
    x_new: pytensor.tensor.TensorVariable
        The cloned variable with the new name
    """
    x_new = x.clone()
    x_new.name = f"{x.name}{suffix}"
    return x_new


def get_required_inputs(graph):
    if not isinstance(graph, list):
        graph = [graph]

    return [
        v
        for v in graph_inputs(graph)
        if isinstance(v, pt.TensorVariable)
        and not isinstance(v, pt.TensorConstant | SharedVariable)
    ]


def rewrite_pregrad(graph):
    """Apply simplifying or stabilizing rewrites to graph that are safe to use
    pre-grad.
    """
    return rewrite_graph(graph, include=("canonicalize", "stabilize"))



from functools import reduce
from typing import cast

import numpy as np
import pytensor
import pytensor.tensor as pt
import pytest
import sympy as sp

from cge_modeling import Equation
from cge_modeling.base.primitives import Variable, _SympyEquation
from cge_modeling.production_functions import (
    BACKEND_TYPE,
    CES,
    _add_second_alpha,
    _check_pairwise_lengths_match,
    cobb_douglass,
    dixit_stiglitz,
    leontief,
    unpack_string_inputs,
)
from cge_modeling.tools.sympy_tools import (
    expand_obj_by_indices,
    substitute_reduce_ops,
)


@pytest.mark.parametrize(
    "dims, coords, expected",
    [
        (None, None, (["L"], ["w"])),
        ("i", {"i": ["A", "B", "C"]}, (["L_A", "L_B", "L_C"], ["w_A", "w_B", "w_C"])),
        (
            ["i", "j"],
            {"i": ["A", "B", "C"], "j": ["D", "E", "F"]},
            (
                [
                    "L_A_D",
                    "L_A_E",
                    "L_A_F",
                    "L_B_D",
                    "L_B_E",
                    "L_B_F",
                    "L_C_D",
                    "L_C_E",
                    "L_C_F",
                ],
                [
                    "w_A_D",
                    "w_A_E",
                    "w_A_F",
                    "w_B_D",
                    "w_B_E",
                    "w_B_F",
                    "w_C_D",
                    "w_C_E",
                    "w_C_F",
                ],
            ),
        ),
    ],
)
def test_unpack_string_inputs(dims, coords, expected):
    factors = "L"
    factor_prices = "w"
    factors, factor_prices = unpack_string_inputs(factors, factor_prices, dims=dims, coords=coords)

    assert factors == expected[0]
    assert factor_prices == expected[1]


def test_add_second_alpha():
    factors = ["L", "K"]
    alpha = _add_second_alpha("alpha", factors)
    assert alpha == ["alpha", "1 - alpha"]


def test_add_second_alpha_doesnt_add():
    # Case 1: single factor
    factors = ["L"]
    alpha = _add_second_alpha("alpha", factors)
    assert alpha == ["alpha"]

    # Case 2: too many factors
    factors = ["L", "K", "M"]
    alpha = _add_second_alpha("alpha", factors)
    assert alpha == ["alpha"]


def test_check_pairwise_lengths():
    factors = ["K"]
    factor_prices = ["r"]
    alpha = ["alpha", "1 - alpha"]
    with pytest.raises(ValueError, match="Lengths of factors and alpha do not match"):
        _check_pairwise_lengths_match(
            ["factors", "factor_prices", "alpha"], [factors, factor_prices, alpha]
        )


def setup_cobb_douglass(alpha):
    factors = ["L", "K"]
    factor_prices = ["w", "r"]
    output = "Y"
    output_price = "P"
    A = "A"

    [eq_production, L_demand, K_demand] = cobb_douglass(
        factors=factors,
        factor_prices=factor_prices,
        output=output,
        output_price=output_price,
        TFP=A,
        factor_shares=alpha,
    )

    return eq_production, L_demand, K_demand


@pytest.mark.parametrize("alpha", ["alpha", ["alpha", "1 - alpha"]], ids=["single", "double"])
def test_cobb_douglass(alpha):
    eq_production, L_demand, K_demand = setup_cobb_douglass(alpha)
    assert eq_production == "Y = A * L ** (alpha) * K ** (1 - alpha)"
    assert L_demand == "L = (alpha) * Y * (P) / (w)"
    assert K_demand == "K = (1 - alpha) * Y * (P) / (r)"


def test_cobb_douglass_pytensor_compile():
    eq_production, L_demand, K_demand = setup_cobb_douglass("alpha")

    def expected_Y(L, K, A, alpha):
        return A * L**alpha * K ** (1 - alpha)

    def expected_L(Y, w, P, alpha):
        return alpha * Y * P / w

    def expected_K(Y, r, P, alpha):
        return (1 - alpha) * Y * P / r

    def compile_Y():
        L = pt.dscalar("L")
        K = pt.dscalar("K")
        A = pt.dscalar("A")
        alpha = pt.dscalar("alpha")

        exec(eq_production)
        return locals()["Y"], [K, L, A, alpha]

    Y, inputs = compile_Y()
    f_Y = pytensor.function(inputs, Y, on_unused_input="ignore", mode="FAST_COMPILE")
    expected = expected_Y(L=1, K=1, A=1, alpha=0.5)
    np.testing.assert_allclose(f_Y(1, 1, 1, 0.5), expected)

    def compile_L():
        Y = pt.dscalar("Y")
        w = pt.dscalar("w")
        P = pt.dscalar("P")
        alpha = pt.dscalar("alpha")

        exec(L_demand)
        return locals()["L"], [Y, w, P, alpha]

    L, inputs = compile_L()
    f_L = pytensor.function(inputs, L, on_unused_input="ignore", mode="FAST_COMPILE")
    expected = expected_L(Y=1, w=1, P=1, alpha=0.5)
    np.testing.assert_allclose(f_L(1, 1, 1, 0.5), expected)

    def compile_K():
        Y = pt.dscalar("Y")
        r = pt.dscalar("r")
        P = pt.dscalar("P")
        alpha = pt.dscalar("alpha")

        exec(K_demand)
        return locals()["K"], [Y, r, P, alpha]

    K, inputs = compile_K()
    f_K = pytensor.function(inputs, K, on_unused_input="ignore", mode="FAST_COMPILE")
    expected = expected_K(Y=1, r=1, P=1, alpha=0.5)
    np.testing.assert_allclose(f_K(1, 1, 1, 0.5), expected)


@pytest.mark.parametrize("alpha", ["alpha", ["alpha", "1 - alpha"]], ids=["single", "double"])
def test_CES(alpha):
    factors = ["L", "K"]
    factor_prices = ["w", "r"]
    output = "Y"
    output_price = "P"
    A = "A"
    epsilon = "epsilon"

    eq_production, *(L_demand, K_demand) = CES(
        factors=factors,
        factor_prices=factor_prices,
        output=output,
        output_price=output_price,
        TFP=A,
        factor_shares=alpha,
        epsilon=epsilon,
    )

    assert eq_production == (
        "Y = A * ((alpha) * L ** ((epsilon - 1) / epsilon) + (1 - alpha) * K ** ((epsilon - 1) / "
        "epsilon)) ** (epsilon / (epsilon - 1))"
    )
    assert L_demand == "L = Y / A * ((alpha) * P * A / (w)) ** epsilon"
    assert K_demand == "K = Y / A * ((1 - alpha) * P * A / (r)) ** epsilon"


def test_computation_of_CES():
    factors = ["L", "K"]
    factor_prices = ["w", "r"]
    output = "Y"
    output_price = "P"
    A = "A"
    epsilon = "epsilon"

    eq_production, *(L_demand, K_demand) = CES(
        factors=factors,
        factor_prices=factor_prices,
        output=output,
        output_price=output_price,
        TFP=A,
        factor_shares="alpha",
        epsilon=epsilon,
    )
    inputs = ["L", "K", "w", "r", "Y", "P", "A", "epsilon", "alpha"]

    eq = sp.parse_expr(eq_production, transformations="all")
    f_eq = sp.lambdify(inputs, eq.rhs)

    def CES_Y(L, K, A, alpha, epsilon):
        return A * (
            (alpha) * L ** ((epsilon - 1) / epsilon) + (1 - alpha) * K ** ((epsilon - 1) / epsilon)
        ) ** (epsilon / (epsilon - 1))

    sympy_out = f_eq(L=10, K=8, w=1, r=1, P=1, A=1, alpha=0.5, epsilon=3, Y=1)
    exact = CES_Y(L=10, K=8, A=1, alpha=0.5, epsilon=3)

    np.testing.assert_allclose(sympy_out, exact)

    def CES_demand(factor_price, output, output_price, TFP, factor_share, epsilon):
        return output / TFP * ((factor_share) * output_price * TFP / factor_price) ** epsilon

    for demand, a_val in zip([L_demand, K_demand], [0.3, 0.7]):
        eq = sp.parse_expr(demand, transformations="all")
        f_eq = sp.lambdify(inputs, eq.rhs)
        sympy_out = f_eq(L=10, K=8, w=1, r=1, P=1, A=2, alpha=0.3, epsilon=3, Y=1)
        exact = CES_demand(1, 1, 1, 2, a_val, 3)
        np.testing.assert_allclose(sympy_out, exact)


@pytest.mark.parametrize("A", ["A", None], ids=["A", "no_A"])
@pytest.mark.parametrize("alpha", ["alpha", None], ids=["alpha", "no_alpha"])
@pytest.mark.parametrize("backend", ["numba", "pytensor"], ids=["numba", "pytensor"])
def test_dixit_stiglitz(A, alpha, backend):
    factors = "X"
    factor_prices = "P_X"
    output = "Y"
    output_price = "P_Y"
    epsilon = "epsilon"

    coords = {"i": ["A", "B", "C"]}
    dims = "i"

    eq_production, X_demand = dixit_stiglitz(
        factors=factors,
        factor_prices=factor_prices,
        output=output,
        output_price=output_price,
        TFP=A,
        factor_shares=alpha,
        epsilon=epsilon,
        dims=dims,
        coords=coords,
        backend=cast(BACKEND_TYPE, backend),
        use_value_definition=False,
    )

    alpha_str = "alpha * " if alpha is not None else ""
    if backend == "numba":
        prod_expected_inner = f"Sum({alpha_str}X ** ((epsilon - 1) / epsilon), ({dims}, 0, 2)) ** (epsilon / (epsilon - 1))"
    elif backend == "pytensor":
        prod_expected_inner = (
            f"({alpha_str}X ** ((epsilon - 1) / epsilon)).sum() ** (epsilon / (epsilon - 1))"
        )
    else:
        assert False

    A_str = "A * " if A is not None else ""
    prod_expected = f"Y = {A_str}{prod_expected_inner}"
    X_demand_expected = f"X = Y / {A_str}({A_str}{alpha_str}P_Y / P_X) ** epsilon"

    assert eq_production == prod_expected
    assert X_demand == X_demand_expected


@pytest.mark.parametrize("backend", ["numba", "pytensor"], ids=["numba", "pytensor"])
def test_dixit_stiglitz_computation(backend):
    def dx_Y(X, A, alpha, epsilon):
        return A * (alpha * X ** ((epsilon - 1) / epsilon)).sum() ** (epsilon / (epsilon - 1))

    def dx_X(Y, A, alpha, P_Y, P_X, epsilon):
        return Y / A * ((alpha * A * P_Y) / P_X) ** epsilon

    factors = "X"
    factor_prices = "P_X"
    output = "Y"
    output_price = "P_Y"
    epsilon = "epsilon"
    dims = "f"
    coords = {"f": ["A", "B", "C", "D", "E", "F", "G"]}
    eq_production, X_demand = dixit_stiglitz(
        factors=factors,
        factor_prices=factor_prices,
        output=output,
        output_price=output_price,
        TFP="A",
        factor_shares="alpha",
        epsilon=epsilon,
        dims=dims,
        coords=coords,
        backend=cast(BACKEND_TYPE, backend),
        use_value_definition=False,
    )

    inputs = ["X", "P_X", "Y", "P_Y", "A", "alpha", "epsilon"]
    f = sp.Idx("f")

    local_dict = {
        "X": sp.IndexedBase("X")[f],
        "P_X": sp.IndexedBase("P_X")[f],
        "alpha": sp.IndexedBase("alpha")[f],
        "Y": sp.Symbol("Y"),
        "P_Y": sp.Symbol("P_Y"),
        "A": sp.Symbol("A"),
        "epsilon": sp.Symbol("epsilon"),
        "f": f,
    }

    if backend == "numba":
        eq = sp.parse_expr(eq_production, transformations="all", local_dict=local_dict)

        f_eq = sp.lambdify(inputs, eq.rhs.doit())
        k = len(coords["f"])

        sympy_out = f_eq(
            X=np.arange(k),
            P_X=np.ones(k),
            Y=1,
            P_Y=1,
            A=1,
            alpha=np.full(k, 1 / k),
            epsilon=3,
        )

        exact = dx_Y(X=np.arange(k), A=1, alpha=np.full(k, 1 / k), epsilon=3)
        np.testing.assert_allclose(sympy_out, exact)

        # Need to remove the index so the lambdify will vectorize
        local_dict.update(
            {
                "X": sp.IndexedBase("X"),
                "P_X": sp.IndexedBase("P_X"),
                "alpha": sp.IndexedBase("alpha"),
            }
        )

        eq = sp.parse_expr(X_demand, transformations="all", local_dict=local_dict)
        f_eq = sp.lambdify(inputs, eq.rhs)

        sympy_out = f_eq(
            X=np.arange(k),
            P_X=np.ones(k),
            Y=1,
            P_Y=1,
            A=1,
            alpha=np.full(k, 1 / k),
            epsilon=3,
        )
        exact = dx_X(Y=1, A=1, alpha=np.full(k, 1 / k), P_Y=1, P_X=np.ones(k), epsilon=3)
        np.testing.assert_allclose(sympy_out, exact)

    elif backend == "pytensor":

        def compile_Y():
            X = pt.dvector("X")
            P_X = pt.dvector("P_X")
            alpha = pt.dvector("alpha")
            P_Y = pt.dscalar("P_Y")
            A = pt.dscalar("A")
            epsilon = pt.dscalar("epsilon")

            exec(eq_production)
            return locals()["Y"], [X, P_X, P_Y, alpha, A, epsilon]

        Y, inputs = compile_Y()
        f_Y = pytensor.function(inputs, Y, on_unused_input="ignore", mode="FAST_COMPILE")
        expected = dx_Y(X=np.arange(7), A=1, alpha=np.full(7, 1 / 7), epsilon=3)
        np.testing.assert_allclose(
            f_Y(np.arange(7), np.ones(7), 1, np.full(7, 1 / 7), 1, 3), expected
        )

        def compile_X():
            P_X = pt.dvector("P_X")
            Y = pt.dscalar("Y")
            P_Y = pt.dscalar("P_Y")
            A = pt.dscalar("A")
            alpha = pt.dvector("alpha")
            epsilon = pt.dscalar("epsilon")

            exec(X_demand)
            return locals()["X"], [Y, P_Y, P_X, A, alpha, epsilon]

        X, inputs = compile_X()
        f_X = pytensor.function(inputs, X, on_unused_input="ignore", mode="FAST_COMPILE")
        expected = dx_X(Y=1, A=1, alpha=np.full(7, 1 / 7), P_Y=1, P_X=np.ones(7), epsilon=3)
        np.testing.assert_allclose(f_X(1, 1, np.ones(7), 1, np.full(7, 1 / 7), 3), expected)


@pytest.mark.parametrize("backend", ["numba", "pytensor"], ids=["numba", "pytensor"])
@pytest.mark.parametrize(
    "coords",
    [
        {"i": ["A", "B", "C"], "j": ["A", "B", "C"]},
        {"i": ["A", "B", "C"], "j": ["E", "F"]},
    ],
    ids=["square", "rectangular"],
)
def test_2d_leontief(backend: BACKEND_TYPE, coords: dict):
    factors = ["X"]
    factor_prices = ["P_X"]
    output = "Y"
    output_price = "P_Y"
    factor_shares = ["phi_X"]

    dims = list(coords.keys())

    zero_profit, *factor_demands = leontief(
        factors=factors,
        factor_prices=factor_prices,
        output=output,
        output_price=output_price,
        factor_shares=factor_shares,
        dims=dims,
        coords=coords,
        backend=backend,
        sum_dim="i",
    )

    if backend == "numba":
        variables = [
            Variable(
                "X",
                dims=("i", "j"),
                description="Demand for good <dim:i> by sector <dim:j>",
            ),
            Variable("P_Y", dims=("i",), description="Price of sector <dim:i> output"),
            Variable("P_X", dims=("i",), description="Price of good <dim:j>"),
            Variable("Y", dims="i", description="Output of sector <dim:i>"),
        ]
        str_dim_to_symbol = {dim: sp.Idx(dim) for dim in coords.keys()}
        local_dict = {x.name: x.to_sympy() for x in variables} | str_dim_to_symbol

        eq = Equation("Zero profit condition for <dim:j>", zero_profit)
        sympy_eq = sp.parse_expr(eq.equation, local_dict=local_dict, transformations="all")
        norm_eq = substitute_reduce_ops(sympy_eq.lhs - sympy_eq.rhs, coords)
        eq = _SympyEquation(
            name=eq.name,
            equation=eq.equation,
            symbolic_eq=sympy_eq,
            _eq=norm_eq,
            _fancy_eq=sympy_eq,
            dims="j",
            eq_id=0,
        )

        eqs = expand_obj_by_indices(eq, coords=coords, dims=None, on_unused_dim="ignore")

        all_atoms = reduce(
            lambda left, right: left.union(right), [eq._eq.atoms() for eq in eqs], set()
        )
        ex_dict = {atom.name: atom for atom in all_atoms if not atom.is_number}
        ex_dict.update({x: sp.IndexedBase(x) for x in ["X", "P_X", "Y", "P_Y"]})

        expected_outputs = [
            f"Y[{dim_2}] - ("
            + " + ".join([f"P_X[{dim_1}] * X[{dim_1}, {dim_2}]" for dim_1 in coords["i"]])
            + f")/P_Y[{dim_2}]"
            for dim_2 in coords["j"]
        ]

        assert len(eqs) == len(coords["j"])

        for eq, ex_eq in zip(eqs, expected_outputs):
            ex_sympy = sp.parse_expr(ex_eq, local_dict=ex_dict, transformations="all")
            assert str(eq._eq) == str(ex_sympy)

    elif backend == "pytensor":
        assert (
            zero_profit
            == f"{output} = ({factor_prices[0]}[:, None] * {factors[0]}).sum(axis=0).ravel() / ({output_price})"
        )
        assert factor_demands[0] == f"{factors[0]} = {factor_shares[0]} * {output}[None]"
    else:
        assert False


def test_1d_leontief():
    factors = ["VA", "VC"]
    factor_prices = ["P_VA", "P_VC"]
    output = "Y"
    output_price = "P_Y"
    factor_shares = ["phi_VA", "phi_VC"]

    coords = {"i": ["A", "B", "C"], "j": ["A", "B", "C"]}
    dims = "i"

    zero_profit, *factor_demands = leontief(
        factors=factors,
        factor_prices=factor_prices,
        output=output,
        output_price=output_price,
        factor_shares=factor_shares,
        dims=dims,
        coords=coords,
        backend="numba",
    )

    assert zero_profit == "Y = (P_VA * VA + P_VC * VC) / (P_Y)"
    for i, demand in enumerate(factor_demands):
        assert demand == f"{factors[i]} = {output} * {factor_shares[i]}"


@pytest.mark.parametrize("backend", ["numba", "pytensor"], ids=["numba", "pytensor"])
def test_leonteif_1d_computation(backend):
    def leontief_Y(VA, VC, P_VA, P_VC, P_Y):
        return (P_VA * VA + P_VC * VC) / P_Y

    def leontief_VA(Y, phi_VA, *args, **kwargs):
        return Y * phi_VA

    def leontief_VC(Y, phi_VC, *args, **kwargs):
        return Y * phi_VC

    factors = ["VA", "VC"]
    factor_prices = ["P_VA", "P_VC"]
    output = "Y"
    output_price = "P_Y"
    factor_shares = ["phi_VA", "phi_VC"]

    coords = {"i": ["A", "B", "C"], "j": ["A", "B", "C"]}
    dims = "i"

    zero_profit, *factor_demands = leontief(
        factors=factors,
        factor_prices=factor_prices,
        output=output,
        output_price=output_price,
        factor_shares=factor_shares,
        dims=dims,
        coords=coords,
        backend=cast(BACKEND_TYPE, backend),
    )

    inputs = ["VA", "VC", "P_VA", "P_VC", "Y", "P_Y", "phi_VA", "phi_VC"]
    local_dict = {
        "VA": sp.IndexedBase("VA"),
        "VC": sp.IndexedBase("VC"),
        "P_VA": sp.IndexedBase("P_VA"),
        "P_VC": sp.IndexedBase("P_VC"),
        "Y": sp.Symbol("Y"),
        "P_Y": sp.Symbol("P_Y"),
        "phi_VA": sp.IndexedBase("phi_VA"),
        "phi_VC": sp.IndexedBase("phi_VC"),
    }

    if backend == "numba":
        eq = sp.parse_expr(zero_profit, transformations="all", local_dict=local_dict)
        f_eq = sp.lambdify(inputs, eq.rhs.doit())
        k = len(coords["i"])

        sympy_out = f_eq(
            VA=np.arange(k),
            VC=np.arange(k),
            P_VA=np.ones(k),
            P_VC=np.ones(k),
            Y=1,
            P_Y=1,
            phi_VA=np.full(k, 1 / k),
            phi_VC=np.full(k, 1 / k),
        )

        exact = leontief_Y(
            VA=np.arange(k), VC=np.arange(k), P_VA=np.ones(k), P_VC=np.ones(k), P_Y=1
        )
        np.testing.assert_allclose(sympy_out, exact)

        for factor_demand, f_expected in zip(factor_demands, [leontief_VA, leontief_VC]):
            eq = sp.parse_expr(factor_demand, transformations="all", local_dict=local_dict)
            f_eq = sp.lambdify(inputs, eq.rhs)
            sympy_out = f_eq(
                VA=np.arange(k),
                VC=np.arange(k),
                P_VA=np.ones(k),
                P_VC=np.ones(k),
                Y=1,
                P_Y=1,
                phi_VA=np.full(k, 1 / k),
                phi_VC=np.full(k, 1 / k),
            )
            exact = f_expected(Y=1, phi_VA=np.full(k, 1 / k), phi_VC=np.full(k, 1 / k))
            np.testing.assert_allclose(sympy_out, exact)

    elif backend == "pytensor":

        def compile_Y():
            VA = pt.dvector("VA")
            VC = pt.dvector("VC")
            P_VA = pt.dvector("P_VA")
            P_VC = pt.dvector("P_VC")
            P_Y = pt.dscalar("P_Y")

            exec(zero_profit)
            return locals()["Y"], [VA, VC, P_VA, P_VC, P_Y]

        Y, inputs = compile_Y()
        f_Y = pytensor.function(inputs, Y, on_unused_input="ignore", mode="FAST_COMPILE")
        expected = leontief_Y(
            VA=np.arange(7), VC=np.arange(7), P_VA=np.ones(7), P_VC=np.ones(7), P_Y=1
        )
        np.testing.assert_allclose(
            f_Y(np.arange(7), np.arange(7), np.ones(7), np.ones(7), 1), expected
        )

        def compile_VA():
            Y = pt.dscalar("Y")
            phi_VA = pt.dvector("phi_VA")

            exec(factor_demands[0])
            return locals()["VA"], [Y, phi_VA]

        VA, inputs = compile_VA()
        f_VA = pytensor.function(inputs, VA, on_unused_input="ignore", mode="FAST_COMPILE")
        expected = leontief_VA(Y=1, phi_VA=np.full(7, 1 / 7))
        np.testing.assert_allclose(f_VA(1, np.full(7, 1 / 7)), expected)

        def compile_VC():
            Y = pt.dscalar("Y")
            phi_VC = pt.dvector("phi_VC")

            exec(factor_demands[1])
            return locals()["VC"], [Y, phi_VC]

        VC, inputs = compile_VC()
        f_VC = pytensor.function(inputs, VC, on_unused_input="ignore", mode="FAST_COMPILE")
        expected = leontief_VC(Y=1, phi_VC=np.full(7, 1 / 7))
        np.testing.assert_allclose(f_VC(1, np.full(7, 1 / 7)), expected)


@pytest.mark.parametrize("backend", ["numba", "pytensor"], ids=["numba", "pytensor"])
@pytest.mark.parametrize(
    "dims, coords",
    [
        (("i", "j"), {"i": ["A", "B", "C"], "j": ["A", "B", "C"]}),
        (("i", "j"), {"i": ["A", "B", "C"], "j": ["D", "E"]}),
    ],
    ids=["square", "rectangular"],
)
def test_2d_leontief_computation(backend, dims, coords):
    def leontief_Y(X, P_X, P_Y):
        return (P_X[:, None] * X).sum(axis=0) / P_Y

    def leontief_X(Y, phi_X):
        return Y * phi_X

    factors = ["X"]
    factor_prices = ["P_X"]
    output = "Y"
    output_price = "P_Y"
    factor_shares = ["phi_X"]

    zero_profit, *factor_demands = leontief(
        factors=factors,
        factor_prices=factor_prices,
        output=output,
        output_price=output_price,
        factor_shares=factor_shares,
        dims=dims,
        coords=coords,
        backend=backend,
        sum_dim="i",
    )

    inputs = ["X", "P_X", "Y", "P_Y", "phi_X"]
    i = sp.Idx("i")
    j = sp.Idx("j")

    local_dict = {
        "X": sp.IndexedBase("X")[i, j],
        "P_X": sp.IndexedBase("P_X")[i],
        "Y": sp.Symbol("Y"),
        "P_Y": sp.Symbol("P_Y"),
        "phi_X": sp.IndexedBase("phi_X")[i, j],
        "i": i,
        "j": j,
    }

    i_len = len(coords["i"])
    j_len = len(coords["j"])

    X_prices = np.random.normal(size=i_len) ** 2
    phi_X_vals = np.random.dirichlet(np.ones(i_len), size=(j_len,)).T
    X_val = np.random.normal(size=(i_len, j_len))
    Y = np.random.normal(size=j_len) ** 2
    P_Y = np.random.normal(size=j_len) ** 2

    if backend == "numba":
        eq = sp.parse_expr(zero_profit, transformations="all", local_dict=local_dict)
        f_eq = sp.lambdify([*inputs, "j"], eq.rhs)

        sympy_out = f_eq(X=X_val, P_X=X_prices, Y=Y, P_Y=P_Y, phi_X=phi_X_vals, j=np.arange(j_len))
        exact = leontief_Y(X=X_val, P_X=X_prices, P_Y=P_Y)
        np.testing.assert_allclose(sympy_out, exact)

        eq = sp.parse_expr(factor_demands[0], transformations="all", local_dict=local_dict)
        f_eq = sp.lambdify([*inputs, "i", "j"], eq.rhs)
        exact = leontief_X(Y=Y, phi_X=phi_X_vals)
        sympy_out = f_eq(
            X=X_val,
            P_X=X_prices,
            Y=Y,
            P_Y=P_Y,
            phi_X=phi_X_vals,
            i=np.arange(i_len)[:, None],
            j=np.arange(j_len)[None],
        )
        np.testing.assert_allclose(sympy_out, exact)

    elif backend == "pytensor":

        def compile_Y():
            P_Y = pt.dvector("P_Y")
            X = pt.dmatrix("X")
            P_X = pt.dvector("P_X")

            exec(zero_profit)
            return locals()["Y"], [P_Y, X, P_X]

        def compile_X():
            Y = pt.dvector("Y")
            phi_X = pt.dmatrix("phi_X")

            exec(factor_demands[0])
            return locals()["X"], [Y, phi_X]

        Y, inputs = compile_Y()
        f_Y = pytensor.function(inputs, Y, on_unused_input="ignore", mode="FAST_COMPILE")
        expected = leontief_Y(X=X_val, P_X=X_prices, P_Y=P_Y)
        np.testing.assert_allclose(f_Y(P_Y, X_val, X_prices), expected)



import re
import warnings

from collections import defaultdict
from copy import deepcopy

import arviz as az
import numpy as np
import sympy as sp
import xarray as xr

from IPython.display import Latex, display
from latextable import draw_latex
from sympy.printing.latex import LatexPrinter
from texttable import Texttable

from cge_modeling.base.primitives import ModelObject
from cge_modeling.base.utilities import flat_array_to_variable_dict
from cge_modeling.tools.sympy_tools import symbol


def get_node_level(node, level=0):
    if node.get_parent() is None:
        return level

    return get_node_level(node.get_parent(), level + 1)


def get_n_descendents(node, total=0):
    if node.n_children == 0:
        return total

    return node.n_children + sum(get_n_descendents(child, 0) for child in node.children)


def get_n_generations(node, total=0):
    if node.n_children == 0:
        return total
    return int(node.n_children > 0) + sum(get_n_generations(child, 0) for child in node.children)


def compute_padding(node):
    pad = max(0, node.n_descendents - 1)
    pad -= sum(max(0, descendent.n_descendents - 1) for descendent in node.get_all_descendents())

    return pad


class Node:
    """
    Class representing a node in a tree data structure. Nodes are connected in parent-child relationships, and
    can also hold data.

    I currently assume that all nodes have only a single parent.

    TODO: Break the coupling between nodes. The child nodes update the parent when created, which is
        not the best design.
    """

    def __init__(self, name, data=None, parent=None):
        self.name = name
        self.parent = parent
        self.children = []
        self.data = data

        self.update_parent()
        self.level = get_node_level(self)

    def add_child(self, child):
        self.children.append(child)

    @property
    def n_children(self):
        return len(self.children)

    @property
    def n_descendents(self):
        return get_n_descendents(self)

    @property
    def n_generations(self):
        return get_n_generations(self)

    def update_parent(self):
        if self.parent is not None:
            self.parent.add_child(self)

    def get_parent(self):
        return self.parent

    def get_all_descendents(self):
        descendents = self.children.copy()

        for child in self.children:
            descendents.extend(child.get_all_descendents())

        return descendents

    def get_all_ancestors(self):
        if self.parent is not None:
            ancestors = [self.parent]
            ancestors.extend(self.parent.get_all_ancestors())
            return ancestors

        else:
            return []

    def __repr__(self):
        return self.name


class Tree:
    """
    Class representing a tree data structure. Trees are built from nodes. The Tree class holds a list of nodes,
    and some helper methods for working with them.
    """

    def __init__(self, nodes=None):
        nodes = [] if nodes is None else nodes
        self.nodes = nodes

    def add_node(self, node):
        if node not in self.nodes:
            self.nodes.append(node)

    @property
    def n_nodes(self):
        return len(self.nodes)

    @property
    def max_level(self):
        return max(node.level for node in self.nodes) + 1

    def get_nodes(self, level):
        return [node for node in self.nodes if node.level == level]

    def get_roots(self):
        return [node for node in self.nodes if node.parent is None]

    def get_leaves(self):
        return [node for node in self.nodes if node.n_children == 0]

    def get_family_trees(self):
        roots = [node for node in self.nodes if node.parent is None]
        return [Tree([root, *root.get_all_descendents()]) for root in roots]

    def get_subtree(self, node):
        family = [node, *node.get_all_descendents()]
        return Tree(family)

    def split_tree(self, cut_level):
        pruned_nodes = []
        update_map = {}
        for level in range(cut_level, self.max_level):
            level_nodes = self.get_nodes(level)
            for old_node in level_nodes:
                new_node = Node(
                    old_node.name,
                    parent=None if level == cut_level else update_map[old_node.parent],
                )
                update_map[old_node] = new_node
                pruned_nodes.append(new_node)

        pruned_tree = Tree(pruned_nodes)
        return pruned_tree

    def make_headings(self):
        """
        Create a list of lists of table headings, where parent nodes are listed above child nodes. In this way,
        the children can be understood as "subheadings".
        """
        headings = defaultdict(list)
        for root in self.get_roots():
            family = self.get_subtree(root)
            for level in range(self.max_level):
                parent_level = family.get_nodes(level - 1)
                family_level = family.get_nodes(level)
                if len(family_level) == 0:
                    n_leaves = len(family.get_leaves())
                    headings[level].extend([None] * n_leaves)
                elif level == 0:
                    for node in family_level:
                        headings[level].extend([node.name] + [None] * compute_padding(node))
                else:
                    for parent in parent_level:
                        if parent.n_children == 0:
                            headings[level].extend([None])
                        else:
                            for node in parent.children:
                                headings[level].extend([node.name] + [None] * compute_padding(node))
        return list(headings.values())


def build_tree(d, tree=None, parent=None):
    """
    Construct a Tree from a dictionary of dictionaries. The dictionary should have the following structure:
        - All keys are strings, representing columns in a table
        - If a column has subcolumns, its value should be a dictionary that follows these rules
        - If a column does not have subcolumns, its value should be a non-dictionary sequence of values
        - All column values must be the same length.

    Parameters
    ----------
    d: dict
        Dictionary of column-subcolumn-data relationships. See the rules above for details.
    tree: Tree
        Current Tree being constructed. Used in recursive parsing of nested dictionaries; you should never need
        to use this argumnet yourself.
    parent: Node
        Current node being recursively parsed. You don't need to declare this yourself.

    Returns
    -------
    Tree
        A tree data-structure representing the table to be created
    """
    if tree is None:
        tree = Tree()
    for k, v in d.items():
        node = Node(k, parent=parent)
        if not isinstance(v, dict):
            node.data = v
        tree.add_node(node)
        if isinstance(v, dict):
            tree = build_tree(v, tree, parent=node)

    return tree


def heading_to_latex(s):
    return r"\textbf{" + s + "}"


def parse_data(x):
    if isinstance(x, ModelObject):
        return x._full_latex_name
    elif isinstance(x, float | int):
        return x
    elif isinstance(x, str):
        if x.isnumeric():
            return x
        else:
            return r"\text{" + x + "}"
    elif isinstance(x, sp.Eq | sp.Symbol | sp.Add | sp.Mul | sp.Expr):
        return LatexPrinter().doprint(x)
    else:
        raise ValueError(f"Unexpected type {type(x)} in table data")


def make_table(info_dict):
    tree = build_tree(info_dict)
    headings = tree.make_headings()
    headings = [
        [heading_to_latex(s) if s is not None else "" for s in heading] for heading in headings
    ]

    leaves = tree.get_leaves()
    n_cols = len(leaves)

    datas = [leaf.data for leaf in leaves]
    data_lens = [len(data) for data in datas]
    assert all([data_len == data_lens[0] for data_len in data_lens])

    n_data = data_lens[0]
    data_rows = [[parse_data(datas[i][row]) for i in range(n_cols)] for row in range(n_data)]

    equation_table = Texttable()
    equation_table.set_cols_align(["c"] * n_cols)
    equation_table.set_cols_valign(["m"] * n_cols)

    equation_table.add_rows(headings + data_rows)
    latex_table = draw_latex(equation_table)

    return latex_table


def make_summary_table(variables, values=None):
    headings = [r"\text{Symbol}", r"\text{Description}"]
    if values is not None:
        headings += [r"\text{Calibrated Value}"]
    heading_row = [headings]

    summary_table = Texttable()
    summary_table.set_cols_align(["c"] * (2 + int(values is not None)))
    summary_table.set_cols_valign(["m"] * (2 + int(values is not None)))

    if values is None:
        data_rows = [[d["latex_name"], "\\text{" + d["description"] + "}"] for d in variables]
    else:
        data_rows = [
            [d["latex_name"], "\\text{" + d["description"] + "}", value]
            for d, value in zip(variables, values)
        ]

    summary_table.add_rows(heading_row + data_rows)
    latex_table = draw_latex(summary_table)

    return latex_table


def make_equation_table(equation_dict):
    headings = [r"", r"\text{Equation}", r"\text{Equation}"]
    heading_row = [headings]

    equation_table = Texttable()
    equation_table.set_cols_align(["c"] * 3)
    equation_table.set_cols_valign(["m"] * 3)

    data_rows = [
        [d["eq_id"], "\\text{" + d["name"] + "}", LatexPrinter().doprint(d["fancy_eq"])]
        for d in equation_dict
    ]

    equation_table.add_rows(heading_row + data_rows)
    latex_table = draw_latex(equation_table)

    return latex_table


def convert_table_to_array(latex_table):
    # Remove the \begin{table} and \end{table} commands
    latex_table = re.sub(r"\\begin{table}", "", latex_table)
    latex_table = re.sub(r"\\end{table}", "", latex_table)

    # Remove \caption and \centering
    latex_table = re.sub(r"\\caption{.*?}", "", latex_table)
    latex_table = re.sub(r"\\centering", "", latex_table)
    latex_table = re.sub(r"\\begin{center}", "", latex_table)
    latex_table = re.sub(r"\\end{center}", "", latex_table)

    # Replace \begin{tabular} and \end{tabular} with \begin{array} and \end{array}
    latex_table = re.sub(r"\\begin{tabular}", r"\\begin{array}", latex_table)
    latex_table = re.sub(r"\\end{tabular}", r"\\end{array}", latex_table)

    return latex_table


def display_latex_table(info_dict):
    table = make_table(info_dict)
    display(Latex(convert_table_to_array(table)))


def display_info_as_table(variable_info):
    variable_table = make_summary_table(variable_info)
    display(Latex(convert_table_to_array(variable_table)))


def display_eqs_as_table(equation_dicts):
    equation_table = make_equation_table(equation_dicts)
    display(Latex(convert_table_to_array(equation_table)))


def sub_info_dicts(info, sub_dict):
    out = []
    for d in info:
        name = d["name"]
        index = d.get("index", ())
        latex_name = d["latex_name"]
        description = d["description"]

        indices_subbed = tuple(x.subs(sub_dict) for x in index)
        latex_name_subbed = latex_name
        description_subbed = description

        for idx, value in sub_dict.items():
            latex_name_subbed = re.sub(
                r"([\^_,{])" + str(idx), r"\g<1>" + str(value), latex_name_subbed
            )
            latex_name_subbed = re.sub(str(idx) + r"}", str(value) + "}", latex_name_subbed)

            description_subbed = re.sub(f" {idx!s}$", f" {value!s}", description_subbed)
            description_subbed = re.sub(f" {idx!s} ", f" {value!s} ", description_subbed)

        out.append(
            {
                "name": name,
                "index": indices_subbed,
                "latex_name": latex_name_subbed,
                "description": description_subbed,
            }
        )
    return out


def get_info(name, value_dict, sectors):
    sectors = [str(x) for x in sectors]
    x = value_dict.get(symbol(name, *sectors))
    if x is None:
        x = value_dict.get(symbol(name))
    return x


def latex_print_equations(equation_info, return_latex=False):
    names, equations, equation_ids = (
        [d.get(attr) for d in equation_info] for attr in ["name", "fancy_eq", "eq_id"]
    )

    tex_output = r"\begin{align}"
    for name, eq, eq_id in zip(names, equations, equation_ids):
        tex_output += r"\text{" + name + r"} &\\"
        tex_output += sp.latex(eq) + r"&\tag{" + str(eq_id) + "}" + r"\\"
    tex_output += r"\end{align}"

    if return_latex:
        return tex_output

    else:
        display(Latex(tex_output))


def list_of_array_to_idata(list_of_arrays: list, cge_model):
    """
    Convert a list of arrays to an xarray dataset

    Parameters
    ----------
    list_of_arrays: list
        A list of numpy arrays, each one corresponding to a variable or parameter of the model, with left-most dimension
        corresponding to the approximation step index. It is assumed that variables are ordered before the parameters,
        and that the variables and parameters are in the same order as in the model.

    cge_model: CGEModel
        CGEModel object

    Returns
    -------
    idata: az.InferenceData
        arviz InferenceData object with two groups -- variables and parameters -- each containing an xarray dataset
        with the same dimensions as the input arrays. Dimensions are labeled using coordinates from the CGEModel object.

    """

    variables = cge_model.variables
    parameters = cge_model.parameters
    n_variables = len(variables)

    coords = cge_model.coords.copy()
    coords.update({"step": range(len(list_of_arrays[0]))})

    xr_var_dict = {
        obj.name: (("step", *obj.dims), list_of_arrays[i]) for i, obj in enumerate(variables)
    }
    xr_param_dict = {
        obj.name: (("step", *obj.dims), list_of_arrays[i + n_variables])
        for i, obj in enumerate(parameters)
    }

    ds_vars = xr.Dataset(xr_var_dict, coords=coords)
    ds_params = xr.Dataset(xr_param_dict, coords=coords)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        idata = az.InferenceData(variables=ds_vars, parameters=ds_params)

    return idata


def optimizer_result_to_idata(res, theta, initial_values, mod):
    coords = deepcopy(mod.coords)
    coords["step"] = [0, 1]
    result = flat_array_to_variable_dict(
        np.r_[res.x, theta], mod.variables + mod.parameters, coords
    )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        optim_var_dict = {
            obj.name: (
                ("step", *obj.dims),
                np.stack([initial_values[obj.name], result[obj.name]]),
            )
            for obj in mod.variables
        }
        # optim_param_dict = {obj.name: (obj.dims, result[obj.name]) for obj in mod.parameters}
        optim_param_dict = {
            obj.name: (
                ("step", *obj.dims),
                np.stack([initial_values[obj.name], result[obj.name]]),
            )
            for obj in mod.parameters
        }
        optim_idata = az.InferenceData(
            variables=xr.Dataset(optim_var_dict, coords=coords),
            parameters=xr.Dataset(optim_param_dict, coords=coords),
        )

    return optim_idata



import pytensor
import pytensor.tensor as pt


def test_prod_to_no_zero_prod():
    x = pt.dvector("x")
    z = pt.prod(x)

    assert not z.owner.op.no_zeros_in_input
    f = pytensor.function([x], z)
    assert any([isinstance(node.op, pt.math.Prod) for node in f.maker.fgraph.toposort()])

    for node in f.maker.fgraph.toposort():
        if isinstance(node.op, pt.math.Prod):
            assert node.op.no_zeros_in_input
            break



from pytensor.tensor.math import Prod
from pytensor.tensor.rewriting.basic import node_rewriter, register_stabilize


@register_stabilize
@node_rewriter([Prod])
def prod_to_no_zero_prod(fgraph, node):
    """
    JAX doesn't support gradient computation in the case where there are zeros in the product. We're allowed to promise
    there will never be zeros, which should always be the case for CGE models. This rewrite makes this promse for any
    product Ops that are in the graph.

    Note that this only affects product reduction Ops, it's not the same as a multiplication.
    """
    if isinstance(node.op, Prod) and not node.op.no_zeros_in_input:
        (x,) = node.inputs
        new_op = Prod(dtype=node.op.dtype, acc_dtype=node.op.dtype, no_zeros_in_input=True)
        return [new_op(x)]



from typing import Literal

import numpy as np

from cge_modeling import CGEModel, Equation, Parameter, Variable


def load_model_1(equation_mode: Literal["numba", "pytensor"] | None = None, **kwargs):
    variable_info = [
        # Firm Variables
        Variable(name="Y", description="Total output of goods"),
        Variable(name="L_d", description="Labor demanded by firms"),
        Variable(name="K_d", description="Capital demanded by firms"),
        # Household Variables
        Variable(name="C", description="Household consumption"),
        Variable(name="income", latex_name="Omega", description="Total household income"),
        # Prices
        Variable(name="r", description="Rental rate of capital"),
        Variable(name="P", description="Price level of the good"),
        # Other
        Variable(name="resid", latex_name=r"varepsilon", description="Walrasian residual"),
    ]

    param_info = [
        # Firm Parameters
        Parameter(name="alpha", description="Share of capital in the production process"),
        Parameter(name="A", description="Total factor productivity"),
        # Exogenous variables
        Parameter(name="L_s", description="Exogenous household labor supply"),
        Parameter(name="K_s", description="Exogenous household capital stock"),
        Parameter(name="w", description="Numeraire wage level"),
    ]

    equations = [
        # Firm Equations
        Equation("Final good production ", "Y = A * K_d ** alpha * L_d ** (1 - alpha)"),
        Equation("Firm demand for capital", "r * K_d = alpha * Y * P"),
        Equation("Firm demand for labor", "w * L_d = (1 - alpha) * Y * P"),
        # Household Equations
        Equation("Household income", "income = w * L_s + r * K_s"),
        Equation("Household budget constraint", "C * P = income"),
        # Market clearning conditions
        Equation("Labor market clearing", "L_s = L_d + resid"),
        Equation("Capital market clearing", "K_s = K_d"),
        Equation("Sector <dim:i> goods market clearing", "Y = C"),
    ]

    return CGEModel(
        variables=variable_info,
        parameters=param_info,
        equations=equations,
        **kwargs,
    )


def calibrate_model_1(L_s, K_s, P, r):
    calib_dict = {"L_s": L_s, "K_s": K_s, "P": P, "r": r}

    # Numeraire
    w = 1

    INC = w * L_s + r * K_s
    C = INC / P
    Y = C
    K_d = K_s
    L_d = L_s

    alpha = r * K_d / Y / P
    A = Y / (K_d**alpha * L_d ** (1 - alpha))
    resid = L_d - L_s

    calib_dict["w"] = w
    calib_dict["income"] = INC
    calib_dict["C"] = C
    calib_dict["Y"] = Y
    calib_dict["K_d"] = K_d
    calib_dict["L_d"] = L_d
    calib_dict["alpha"] = alpha
    calib_dict["A"] = A
    calib_dict["resid"] = resid

    return calib_dict


model_1_data = {"L_s": 7000.0, "K_s": 4000.0, "P": 1.0, "r": 1.0}


def expected_model_1_jacobian(Y, L_d, K_d, C, income, r, P, resid, alpha, A, L_s, K_s, w):
    d1_dL_d = -A * K_d**alpha * (1 - alpha) * L_d ** (-alpha)
    d1_dK_d = -A * alpha * K_d ** (alpha - 1) * L_d ** (1 - alpha)

    # @formatter:off
    #                Y,                 L_d,    K_d,      C, inc, r,    P,                 resid
    return np.array(
        [
            [1, d1_dL_d, d1_dK_d, 0, 0, 0, 0, 0],
            [-alpha * P, 0, r, 0, 0, K_d, -alpha * Y, 0],
            [-(1 - alpha) * P, w, 0, 0, 0, 0, -(1 - alpha) * Y, 0],
            [0, 0, 0, 0, 1, -K_s, 0, 0],
            [0, 0, 0, P, -1, 0, C, 0],
            [0, -1, 0, 0, 0, 0, 0, -1],
            [0, 0, -1, 0, 0, 0, 0, 0],
            [1, 0, 0, -1, 0, 0, 0, 0],
        ]
    )
    # @formatter:on


def load_model_2(equation_mode: Literal["numba", "pytensor"] | None = None, **kwargs):
    backend = kwargs.get("backend", "numba")
    if equation_mode is None:
        equation_mode = backend

    sectors = ["0", "1", "2"]
    coords = {"i": sectors, "j": sectors}
    n_sectors = len(sectors)

    variable_info = [
        # Firm Variables
        Variable(name="Y", dims="i", description="Total output of good <dim:i>"),
        Variable(
            name="VA",
            dims="i",
            description="Labor-capital bundle produced by sector <dim:i>",
        ),
        Variable(
            name="VC",
            dims="i",
            description="Intermediate goods bundle produced by sector <dim:i>",
        ),
        Variable(
            name="X",
            dims=("i", "j"),
            description="Demand for sector <dim:i> goods by sector <dim:j>",
        ),
        Variable(
            name="L_d",
            dims="i",
            extend_subscript=True,
            description="Labor demanded by sector <dim:i>",
        ),
        Variable(
            name="K_d",
            dims="i",
            extend_subscript=True,
            description="Capital demanded by sector <dim:i>",
        ),
        # Household Variables
        Variable(name="U", description="Household utility"),
        Variable(name="C", dims="i", description="Household consumption of good <dim:i>"),
        Variable(name="income", latex_name="Omega", description="Total household income"),
        # Prices
        Variable(name="r", description="Rental rate of capital"),
        Variable(name="w", description="Wage level"),
        Variable(
            name="P_VA",
            dims="i",
            extend_subscript=True,
            description="Price of the labor-capital bundle in sector <dim:i>",
        ),
        Variable(
            name="P_VC",
            dims="i",
            extend_subscript=True,
            description="Price of the intermediate bundle in sector <dim:i>",
        ),
        Variable(
            name="P",
            dims="i",
            description="Price level of final goods in sector <dim:i>",
        ),
        # Other
        Variable(name="resid", latex_name=r"varepsilon", description="Walrasian residual"),
    ]

    param_info = [
        # Firm Parameters
        Parameter(
            name="psi_VA",
            extend_subscript=True,
            dims="i",
            description="Share of labor-capital in sector <dim:i>'s final good",
        ),
        Parameter(
            name="psi_VC",
            extend_subscript=True,
            dims="i",
            description="Share of intermediate goods bundle in sector <dim:i>'s final product",
        ),
        Parameter(
            name="psi_X",
            extend_subscript=True,
            dims=("i", "j"),
            description="Share of sector <dim:i>'s final good in sector <dim:j>'s value chain bundle",
        ),
        Parameter(
            name="alpha",
            dims="i",
            description="Share of capital in sector <dim:i>'s production process",
        ),
        Parameter(
            name="phi_VA",
            extend_subscript=True,
            dims="i",
            description="Elasticity of subsitution between input factors in <dim:i> sector VA bundle",
        ),
        Parameter(
            name="A",
            dims="i",
            description="Total factor productivity in sector <dim:i>",
        ),
        # Household Parameters
        Parameter(
            name="gamma",
            dims="i",
            description="Household utility weight on consumption of sector <dim:i> goods",
        ),
        # Exogenous variables
        Parameter(name="L_s", description="Exogenous household labor supply"),
        Parameter(name="K_s", description="Exogenous household capital stock"),
        Parameter(
            name="P_Ag_bar",
            latex_name=r"\bar{P}_{Ag}",
            description="Exogenous agricultural price level",
        ),
    ]

    equations = [
        # # Firm Equations
        # Final Goods
        Equation("Final good production of sector <dim:i>", "P * Y = P_VC * VC + P_VA * VA"),
        Equation("Sector <dim:i> demand for intermediate goods bundle", "VC = psi_VC * Y"),
        Equation("Sector <dim:i> demand for labor-capital", "VA = psi_VA * Y"),
        # Value chain bundle
        Equation(
            "Sector <dim:i> production of intermediate goods bundle",
            "VC * P_VC = (P[:, None] * X).sum(axis=0).ravel()"
            if equation_mode == "pytensor"
            else "VC * P_VC = Sum(P.subs({i:j}) * X.subs([(i,k), (j,i), (k,j)]), "
            + f"(j, 0, {n_sectors - 1}))",
        ),
        Equation(
            "Sector <dim:i> demand for sector <dim:j> intermediate input",
            "X = psi_X * VC[None]" if equation_mode == "pytensor" else "X = psi_X * VC.subs({i:j})",
        ),
        # Value add bundle
        Equation(
            "Sector <dim:i> production of labor-capital",
            "VA = A * (alpha * K_d**((phi_VA - 1) / phi_VA) +"
            "(1 - alpha) * L_d**((phi_VA - 1) / phi_VA)) ** (phi_VA / (phi_VA - 1))",
        ),
        Equation(
            "Sector <dim:i> demand for capital",
            "K_d = VA / A * (alpha * P_VA * A / r) ** phi_VA",
        ),
        Equation(
            "Sector <dim:i> demand for labor",
            "L_d = VA / A * ((1 - alpha) * A * P_VA / w) ** phi_VA",
        ),
        # # Household Equations
        Equation("Household income", "income = w * L_s + r * K_s"),
        Equation(
            "Household utility",
            "U = (C**gamma).prod()"
            if equation_mode == "pytensor"
            else "U = Product(C**gamma, " + f"(i, 0, {n_sectors - 1}))",
        ),
        Equation("Household demand for good <dim:i>", "C = gamma * income / P"),
        # # Market clearning conditions
        Equation(
            "Labor market clearing",
            "L_s = L_d.sum() + resid"
            if equation_mode == "pytensor"
            else "L_s = resid + Sum(L_d, " + f"(i, 0, {n_sectors - 1}))",
        ),
        Equation(
            "Capital market clearing",
            "K_s = K_d.sum()"
            if equation_mode == "pytensor"
            else f"K_s = Sum(K_d, (i, 0, {n_sectors - 1}))",
        ),
        Equation(
            "Sector <dim:i> goods market clearing",
            "Y = C + X.sum(axis=1)"
            if equation_mode == "pytensor"
            else f"Y = C + Sum(X, (j, 0, {n_sectors - 1}))",
        ),
        Equation(
            "Numeraire",
            "P[0] = P_Ag_bar" if equation_mode == "pytensor" else "P.subs({i:0}) = P_Ag_bar",
        ),
    ]

    return CGEModel(
        variables=variable_info,
        parameters=param_info,
        equations=equations,
        coords=coords,
        **kwargs,
    )


def calibrate_model_2(L_d, K_d, Y, X, P, P_VA, P_VC, r, w, phi_VA, P_Ag_bar):
    calib_dict = {
        "L_d": L_d,
        "K_d": K_d,
        "Y": Y,
        "X": X,
        "P": P,
        "r": r,
        "w": w,
        "P_VA": P_VA,
        "P_VC": P_VC,
        "phi_VA": phi_VA,
        "P_Ag_bar": P_Ag_bar,
    }

    rho_VA = (phi_VA - 1) / phi_VA

    # Numeraire
    resid = 0.0

    # Household calibration
    L_s = L_d.sum()
    K_s = K_d.sum()
    income = w * L_s + r * K_s
    C = Y - X.sum(axis=1)
    gamma = C / income * P
    U = (C**gamma).prod()

    # Firm calibration
    VA = (w * L_d + r * K_d) / P_VA
    VC = (P[:, None] * X).sum(axis=0) / P_VC

    alpha = r * K_d ** (1 / phi_VA) / (r * K_d ** (1 / phi_VA) + w * L_d ** (1 / phi_VA))
    A = VA * (alpha * K_d**rho_VA + (1 - alpha) * L_d**rho_VA) ** (-1 / rho_VA)

    psi_VA = VA / Y
    psi_VC = VC / Y
    psi_X = X / VC[None]

    calib_dict["VA"] = VA
    calib_dict["VC"] = VC
    calib_dict["psi_VC"] = psi_VC
    calib_dict["psi_VA"] = psi_VA
    calib_dict["psi_X"] = psi_X
    calib_dict["alpha"] = alpha
    calib_dict["A"] = A

    calib_dict["income"] = income
    calib_dict["C"] = C
    calib_dict["U"] = U
    calib_dict["gamma"] = gamma

    calib_dict["K_s"] = K_s
    calib_dict["L_s"] = L_s
    calib_dict["resid"] = resid
    calib_dict["w"] = w

    return calib_dict


model_2_data = {
    "L_d": np.array([1000.0, 2000.0, 4000.0]),
    "K_d": np.array([500.0, 2000.0, 500.0]),
    "X": np.array([[1000.0, 1000.0, 1000.0], [2000.0, 3500.0, 3000.0], [500.0, 2500.0, 1000.0]]),
    "Y": np.array([5000.0, 11000.0, 9500.0]),
    "phi_VA": np.array([3.0, 3.0, 3.0]),
    "P": np.array([1.0, 1.0, 1.0]),
    "P_VA": np.array([1.0, 1.0, 1.0]),
    "P_VC": np.array([1.0, 1.0, 1.0]),
    "r": 1.0,
    "w": 1.0,
    "P_Ag_bar": 1.0,
}


def expected_model_2_jacobian(
    Y,
    VA,
    VC,
    X,
    L_d,
    K_d,
    U,
    C,
    income,
    r,
    w,
    P_VA,
    P_VC,
    P,
    resid,
    psi_VA,
    psi_VC,
    psi_X,
    alpha,
    A,
    L_s,
    K_s,
    gamma,
    P_Ag_bar,
    phi_VA,
):
    Y_columns = np.array([0, 1, 2])
    VA_columns = np.array([3, 4, 5])
    VC_columns = np.array([6, 7, 8])
    X_columns = np.array([9, 10, 11, 12, 13, 14, 15, 16, 17])
    L_d_columns = np.array([18, 19, 20])
    K_d_columns = np.array([21, 22, 23])
    U_columns = np.array([24])
    C_columns = np.array([25, 26, 27])
    income_columns = np.array([28])
    r_columns = np.array([29])
    w_columns = np.array([30])
    P_VA_columns = np.array([31, 32, 33])
    P_VC_columns = np.array([34, 35, 36])
    P_columns = np.array([37, 38, 39])
    resid_columns = np.array([40])

    J = np.zeros((41, 41))

    # Equation 0-2: P * Y = P_VC * VC + P_VA * VA
    row_idx = np.array([0, 1, 2])
    J[row_idx, P_columns] = Y
    J[row_idx, Y_columns] = P
    J[row_idx, P_VC_columns] = -VC
    J[row_idx, VC_columns] = -P_VC

    J[row_idx, P_VA_columns] = -VA
    J[row_idx, VA_columns] = -P_VA

    # Equation 3-5: VC = psi_VC * Y
    row_idx = np.array([3, 4, 5])
    J[row_idx, VC_columns] = 1
    J[row_idx, Y_columns] = -psi_VC

    # Equation 6-8: VA = psi_VA * Y
    row_idx = np.array([6, 7, 8])
    J[row_idx, VA_columns] = 1
    J[row_idx, Y_columns] = -psi_VA

    # Equation 9-11: VC * P_VC = (P[:, None] * X).sum(axis=0).ravel()
    row_idx = np.array([9, 10, 11])
    row_slice = np.s_[9:12]
    J[row_idx, VC_columns] = P_VC
    J[row_idx, P_VC_columns] = VC
    J[row_slice, X_columns] = -np.hstack([np.eye(3) * P[i] for i in range(3)])
    J[row_slice, P_columns] = -X.T

    # Equation 12-20: X = psi_X * VC[None]
    row_idx = np.array([12, 13, 14, 15, 16, 17, 18, 19, 20])
    J[row_idx, X_columns] = 1

    row_slice = np.s_[12:21]
    col_slice = np.s_[6:9]
    J[row_slice, col_slice] = -np.vstack([np.diag(psi_X[i]) for i in range(3)])

    # Equation 21-23: VA = A * (alpha * K_d**((phi_VA - 1) / phi_VA) + (1 - alpha) * L_d**((phi_VA - 1) / phi_VA)) **
    #                           (phi_VA / (phi_VA - 1))
    row_idx = np.array([21, 22, 23])
    J[row_idx, VA_columns] = 1
    J[row_idx, K_d_columns] = (
        -A
        * (alpha * K_d ** ((phi_VA - 1) / phi_VA) + (1 - alpha) * L_d ** ((phi_VA - 1) / phi_VA))
        ** (1 / (phi_VA - 1))
        * alpha
        * K_d ** (-1 / phi_VA)
    )
    J[row_idx, L_d_columns] = (
        -A
        * (alpha * K_d ** ((phi_VA - 1) / phi_VA) + (1 - alpha) * L_d ** ((phi_VA - 1) / phi_VA))
        ** (1 / (phi_VA - 1))
        * (1 - alpha)
        * L_d ** (-1 / phi_VA)
    )

    # Equation 24-26: K_d = VA / A * (alpha * P_VA * A / r) ** phi_VA
    row_idx = np.array([24, 25, 26])
    J[row_idx, K_d_columns] = 1
    J[row_idx, VA_columns] = -1 / A * (alpha * P_VA * A / r) ** phi_VA
    J[row_idx, P_VA_columns] = (
        -VA / A * (alpha * P_VA * A / r) ** (phi_VA - 1) * alpha * A / r * phi_VA
    )
    J[row_idx, r_columns] = (
        VA / A * (alpha * P_VA * A / r) ** (phi_VA - 1) * alpha * A * phi_VA * P_VA / r**2
    )

    # Equation 27-29: L_d = VA / A * ((1 - alpha) * A * P_VA / w) ** phi_VA
    row_idx = np.array([27, 28, 29])
    J[row_idx, L_d_columns] = 1
    J[row_idx, VA_columns] = -1 / A * ((1 - alpha) * A * P_VA / w) ** phi_VA
    J[row_idx, P_VA_columns] = (
        -VA / A * ((1 - alpha) * A * P_VA / w) ** (phi_VA - 1) * (1 - alpha) * A / w * phi_VA
    )
    J[row_idx, w_columns] = (
        VA
        / A
        * ((1 - alpha) * A * P_VA / w) ** (phi_VA - 1)
        * (1 - alpha)
        * A
        * phi_VA
        * P_VA
        / w**2
    )

    # Equation 30: income = w * L_s + r * K_s
    row_idx = np.array([30])
    J[row_idx, income_columns] = 1
    J[row_idx, w_columns] = -L_s
    J[row_idx, r_columns] = -K_s

    # Equation 31: U = (C**gamma).prod()
    row_idx = np.array([31])
    J[row_idx, U_columns] = 1
    U_val = np.prod(C**gamma)
    J[row_idx, C_columns] = -gamma * U_val / C

    # Equation 32-34: C = gamma * income / P
    row_idx = np.array([32, 33, 34])
    J[row_idx, C_columns] = 1
    J[row_idx, income_columns] = -gamma / P
    J[row_idx, P_columns] = gamma * income / P**2

    # Equation 35: L_s = L_d.sum() + resid
    row_idx = np.array([35])
    J[row_idx, L_d_columns] = -1
    J[row_idx, resid_columns] = -1

    # Equation 36: K_s = K_d.sum()
    row_idx = np.array([36])
    J[row_idx, K_d_columns] = -1

    # Equation 37-39: Y = C + X.sum(axis=1)
    row_idx = np.array([37, 38, 39])
    J[row_idx, Y_columns] = 1
    J[row_idx, C_columns] = -1

    row_slice = np.s_[37:40]
    col_slice = np.s_[9:18]
    J[row_slice, col_slice] = -np.kron(np.eye(3), np.ones(3))

    # Equation 40: P[0] = P_Ag_bar
    row_idx = np.array([40])
    J[row_idx, P_columns[0]] = 1

    return J



import numpy as np
import pytensor
import pytensor.tensor as pt

from numpy.testing import assert_allclose

from cge_modeling.base.utilities import flat_array_to_variable_dict
from cge_modeling.pytensorf.compile import (
    compile_cge_model_to_pytensor_Op,
)
from cge_modeling.pytensorf.optimize import root
from tests.utilities.models import (
    calibrate_model_2,
    load_model_1,
    load_model_2,
    model_2_data,
)

FLOATX = pytensor.config.floatX


def _postprocess_root_return(root_histories: list[pt.TensorVariable]) -> np.ndarray:
    final_symbolic = pt.concatenate(
        [pt.atleast_1d(root_history[-1]).ravel() for root_history in root_histories]
    )
    final_roots = final_symbolic.eval()
    return final_roots


def test_simple_root():
    x = pt.dscalar("x")
    y = pt.dscalar("y")
    eq1 = x**2 + y - 1
    eq2 = x - y**2 + 1
    system = pt.stack([eq1, eq2])
    jac = pt.stack(pytensor.gradient.jacobian(system, [x, y]))

    f = pytensor.compile.builders.OpFromGraph([x, y], [system])
    f_jac = pytensor.compile.builders.OpFromGraph([x, y], [jac])

    x_val = 1.0
    y_val = 1.0

    root_histories, converged, step_size, n_steps = root(f, f_jac, {"x": x_val, "y": y_val})
    final_root = _postprocess_root_return(root_histories)

    assert_allclose(final_root, [0.0, 1.0], atol=1e-8)


def test_small_model():
    def f_model(Y, C, L_d, K_d, P, r, resid, K_s, L_s, A, alpha, w):
        equations = pt.stack(
            [
                Y - A * K_d**alpha * L_d ** (1 - alpha),
                r * K_d - alpha * Y * P,
                w * L_d - (1 - alpha) * Y * P,
                Y - C,
                P * C - w * L_s - r * K_s,
                K_d - K_s,
                L_d - L_s + resid,
            ]
        )
        return equations

    variables = list(map(pt.dscalar, ["Y", "C", "L_d", "K_d", "P", "r", "resid"]))
    params = list(map(pt.dscalar, ["K_s", "L_s", "A", "alpha", "w"]))

    equations = f_model(*variables, *params)
    jac = pt.stack(pytensor.gradient.jacobian(equations, variables)).T
    f_jac = pytensor.compile.builders.OpFromGraph(variables + params, outputs=[jac], inline=True)

    x0 = {"Y": 11000, "C": 11000, "L_d": 7000, "K_d": 4000, "r": 1, "P": 1, "resid": 0}
    param_vals = {"K_s": 4000, "L_s": 7000, "A": 2, "alpha": 0.33, "w": 1}

    root_histories, converged, step_size, n_steps = root(
        f_model, f_jac, initial_data=x0, parameters=param_vals
    )
    root_eval = _postprocess_root_return(root_histories)

    assert converged[-1].eval()
    expected_root = np.array(
        [1.16392629e04, 1.16392629e04, 7000.0, 4000.0, 0.897630824, 0.861940299, 0.0]
    )
    f_model_compiled = pytensor.function(
        inputs=variables + params, outputs=equations, mode="FAST_COMPILE"
    )

    # Check optimizer converges to the scipy result
    assert_allclose(root_eval, expected_root, atol=1e-8)

    # Check Y = C
    assert_allclose(root_eval[0], root_eval[1], atol=1e-8)

    # Check residuals are zero at the root
    assert_allclose(f_model_compiled(*root_eval, **param_vals).ravel(), np.zeros(7), atol=1e-8)


def test_small_model_from_compile():
    mod = load_model_1(parse_equations_to_sympy=False, backend="pytensor", compile=False)
    f_model, f_jac = compile_cge_model_to_pytensor_Op(mod)
    data = {
        "Y": 11000.0,
        "C": 11000.0,
        "income": 11000.0,
        "L_d": 7000.0,
        "K_d": 4000.0,
        "P": 1.0,
        "r": 1.0,
        "resid": 0.0,
    }
    param_data = {"K_s": 4000, "L_s": 7000, "A": 2, "alpha": 0.33, "w": 1}
    expected_roots = {
        "Y": 11639.2629,
        "C": 11639.2629,
        "income": 10447.761194,
        "L_d": 7000.0,
        "K_d": 4000.0,
        "P": 0.897630824,
        "r": 0.861940299,
        "resid": 0.0,
    }

    sorted_data = {k: data[k] for k in mod.variable_names}
    sorted_params = {k: param_data[k] for k in mod.parameter_names}

    root_histories, converged, step_size, n_steps = root(
        f_model,
        f_jac,
        initial_data=sorted_data,
        parameters=sorted_params,
        tol=1e-8,
        max_iter=500,
    )

    root_eval = _postprocess_root_return(root_histories)
    assert converged[-1].eval()

    # Check optimizer converges to the scipy result
    assert_allclose(
        root_eval,
        np.array([expected_roots[var] for var in mod.variable_names], dtype=FLOATX),
        atol=1e-8,
    )

    # Check Y = C
    Y_idx = mod.variable_names.index("Y")
    C_idx = mod.variable_names.index("C")
    assert_allclose(root_eval[Y_idx], root_eval[C_idx], atol=1e-8)

    # Check residuals are zero at the root
    params = np.concatenate([np.atleast_1d(param_data[param]) for param in mod.parameter_names])
    assert_allclose(f_model(*root_eval, *params).eval(), np.zeros(8), atol=1e-8)


def test_sector_model_from_compile():
    mod = load_model_2(parse_equations_to_sympy=False, mode="FAST_COMPILE", backend="pytensor")
    calib_dict = calibrate_model_2(**model_2_data)

    x0 = {var.name: calib_dict[var.name] for var in mod.variables}
    params = {param.name: calib_dict[param.name] for param in mod.parameters}

    f_model, f_jac = compile_cge_model_to_pytensor_Op(mod)
    root_history, converged, step_size, n_steps = root(
        f_model, f_jac, initial_data=x0, parameters=params, tol=1e-8, max_iter=500
    )

    root_eval = _postprocess_root_return(root_history)
    root_point = flat_array_to_variable_dict(root_eval, mod.variables, mod.coords)

    # Check residuals are zero at the root
    assert_allclose(
        mod.f_system(**root_point, **params).ravel(),
        np.zeros(mod.n_variables),
        atol=1e-8,
    )



import logging

_log = logging.getLogger(__name__)
if not logging.root.handlers:
    _log.setLevel(logging.INFO)
    if len(_log.handlers) == 0:
        handler = logging.StreamHandler()
        _log.addHandler(handler)



import pytest

from cge_modeling.plotting import plot_kateplot, plot_lines
from tests.utilities.models import calibrate_model_2, load_model_2, model_2_data


@pytest.fixture(scope="session")
def test_data():
    mod = load_model_2()
    calibrated_data = calibrate_model_2(**model_2_data)
    idata = mod.simulate(calibrated_data, final_delta_pct={"L_s": 0.5})

    return idata, mod, calibrated_data


def test_plot_kateplot(test_data: tuple):
    plot_data, mod, calibrated_data = test_data
    plot_kateplot(
        plot_data,
        mod=mod,
        initial_values=calibrated_data,
        shock_name="L_s",
        var_names="C",
    )


def test_plot_lines(test_data: tuple):
    plot_data, mod, calibrated_data = test_data
    plot_lines(plot_data, mod=mod)



import logging

from cge_modeling.base.cge import CGEModel
from cge_modeling.base.primitives import Equation, Parameter, Variable
from cge_modeling.plotting import plot_kateplot, plot_lines
from cge_modeling.pytensorf.rewrites import prod_to_no_zero_prod  # noqa: F401
from cge_modeling.tools.output_tools import display_info_as_table, latex_print_equations

_log = logging.getLogger("cge_modeling")
if not logging.root.handlers:
    _log.setLevel(logging.INFO)
    if len(_log.handlers) == 0:
        handler = logging.StreamHandler()
        _log.addHandler(handler)


__version__ = "0.0.3"

__all__ = [
    "CGEModel",
    "Variable",
    "Parameter",
    "Equation",
    "latex_print_equations",
    "display_info_as_table",
    "plot_lines",
    "plot_kateplot",
]



import re
import time

from abc import ABC
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime

import pandas as pd
import sympy as sp

from sympy.abc import greeks
from sympy.printing.latex import latex

greeks = [*list(greeks), "varepsilon"]
Greeks = [x.title() for x in greeks]
GREEK_PATTERN = "|".join(
    [f"(((^{greek})|((?=[_^]){greek})|((?<=[_^]){greek})))" for greek in greeks + Greeks]
)


def _pretty_print_dim_flags(s, dims, dim_vals):
    for dim in dims:
        dim_val = dim_vals.get(dim, None)
        if dim_val is None:
            continue
        pattern = f"<dim:{dim!s}>"
        s = re.sub(pattern, str(dim_val), s)
    return s


def _process_subscript_chunks(chunks):
    return ", ".join(chunks)


@dataclass(slots=True, order=True, frozen=False, repr=False)
class ModelObject(ABC):
    name: str
    dims: tuple[str, ...] | list[str, ...] | str | None = None
    dim_vals: dict[str, str] | None = None
    latex_name: str | None = None
    _full_latex_name: str | None = None
    description: str | None = None
    base_name: str = None
    real: bool = True
    positive: bool = True
    extend_subscript: bool | int = False
    assumptions: dict = field(default_factory=dict)

    def __post_init__(self):
        self.assumptions["real"] = self.real
        self.assumptions["positive"] = self.positive

        object.__setattr__(self, "sort_index", self.name)
        object.__setattr__(self, "dims", self._initialize_index())
        object.__setattr__(self, "dim_vals", self._initialize_dim_vals())
        object.__setattr__(self, "base_name", self.name)
        object.__setattr__(self, "latex_name", self._initialize_latex_name())
        object.__setattr__(self, "_full_latex_name", self._initialize_full_latex_name())
        object.__setattr__(self, "description", self._initialize_description())

    def _make_latex_subscript(self):
        idx_strs = []
        for dim in self.dims:
            dim_val = self.dim_vals.get(dim, None)
            dim_val_text = "" if dim_val is None else "=\\text{" + str(dim_val) + "}"
            idx_strs.append(f"{dim}{dim_val_text}")

        if len(idx_strs) == 0:
            return ""

        return "_{" + ", ".join(idx_strs) + "}"

    def _initialize_index(self):
        dims = self.dims

        # If dims is None that's fine, replace it with an empty tuple
        if dims is None:
            return ()

        # Case 1: Index is empty (valid), or a tuple of strings (valid)
        if isinstance(dims, tuple | list):
            if len(dims) == 0:
                return dims
            # Convert to a tuple
            dims = tuple(dims)
            if all([isinstance(x, str) for x in dims]):
                return dims
            else:
                raise ValueError(f"dims must be a string or a tuple of strings, found {dims}")

        # Case 2: Index is a single comma delimited string. Convert to a tuple
        elif isinstance(dims, str):
            dims = dims.split(",")
            return tuple(dim.strip() for dim in dims)

        else:
            raise ValueError(f"dims must be a string or a tuple of strings, found {dims}")

    def _initialize_dim_vals(self):
        dims = self.dims
        dim_vals = self.dim_vals

        # The default should be an empty dictionary
        if dim_vals is None:
            return {}

        elif not isinstance(dim_vals, dict):
            raise ValueError(
                f"dim_vals of {self.base_name} should be a dictionary mapping dimensions to specific coord"
                f" values; found {type(dim_vals)}"
            )

        keys = list(dim_vals.keys())

        # Check all the keys are in the dims
        extra_dims = set(keys) - set(dims)
        if len(extra_dims) > 0:
            raise ValueError(
                f"All dim_vals of {self.base_name} must be associated with known dimensions: {self.dims}. "
                f"Found unknown dimensions: {extra_dims}."
            )
        return dim_vals

    def _initialize_latex_name(self):
        latex_name = self.latex_name if self.latex_name is not None else self.name
        latex_name = re.sub(GREEK_PATTERN, r"\\" + r"\g<0>", latex_name)
        return latex_name

    def _initialize_full_latex_name(self):
        idx_subscript = self._make_latex_subscript()
        base_latex_name = self.latex_name
        int_extend = int(self.extend_subscript)

        if self.extend_subscript:
            tokens = base_latex_name.split("_")
            base_chunks, subscript_chunks = tokens[:-int_extend], tokens[-int_extend:]
            subscript = _process_subscript_chunks(subscript_chunks)

            if len(base_chunks) == 0:
                base_latex_name, subscript = subscript, base_chunks
            else:
                base_latex_name = "_".join(base_chunks)

            if len(subscript) > 0:
                subscript = subscript.replace("{", "").replace("}", "")
                idx_subscript = "_{" + subscript + ", " + idx_subscript[2:]

        latex_name = base_latex_name + idx_subscript

        return latex_name

    def _initialize_description(self):
        if self.description is not None:
            self.description = _pretty_print_dim_flags(self.description, self.dims, self.dim_vals)
            return self.description
        return f"{self._full_latex_name}, Positive = {self.positive}, Real = {self.real}"

    def __getitem__(self, item: str):
        return getattr(self, item)

    def __repr__(self):
        return self._full_latex_name

    def to_dict(self):
        return {
            "name": self.name,
            "dims": self.dims,
            "sympy": self.to_sympy(long_name=False),
            "latex_name": self._full_latex_name,
            "dim_vals": self.dim_vals,
            "description": self.description,
        }

    def to_sympy(self, long_name=False, use_latex_name=False):
        indices = [sp.Idx(dim) for dim in self.dims]
        if use_latex_name:
            base_name = self._full_latex_name if long_name else self.latex_name
        else:
            base_name = self.name if long_name else self.base_name
        if len(indices) == 0:
            return sp.Symbol(base_name, **self.assumptions)
        base = sp.IndexedBase(base_name, **self.assumptions)
        dim_val_subs = {sp.Idx(k): v for k, v in self.dim_vals.items()}

        return base[indices].subs(dim_val_subs)

    def update_dim_value(self, dim, value):
        if dim not in self.dims:
            raise ValueError(f"{dim} is not a valid dimension of {self.name}.")

        self.dim_vals[dim] = value
        self._full_latex_name = self._initialize_full_latex_name()
        self.description = self._initialize_description()

    def copy(self):
        return deepcopy(self)


@dataclass(order=True, frozen=False, repr=False)
class Variable(ModelObject):
    pass


@dataclass(order=True, frozen=False, repr=False)
class Parameter(ModelObject):
    pass


@dataclass(frozen=True)
class Equation:
    name: str
    equation: str
    eq_id: int | None = None

    def _set_eq_id(self, new_id):
        object.__setattr__(self, "eq_id", new_id)

    def __getitem__(self, item: str):
        return getattr(self, item)

    def __repr__(self):
        return self.name


# noinspection PyDataclass
@dataclass(frozen=True, kw_only=True)
class _SympyEquation(Equation):
    symbolic_eq: sp.Eq
    _eq: sp.Expr
    _fancy_eq: sp.Eq
    dims: tuple[str, ...]
    dim_vals: dict[str, str] = field(default_factory=dict)

    def __repr__(self):
        return self.name

    def _repr_latex_(self):
        s = latex(self.symbolic_eq, mode="plain")
        return f"$\\displaystyle {s}$"

    def copy(self):
        return deepcopy(self)

    def update_dim_value(self, dim, value):
        if dim not in self.dims:
            raise ValueError(f"{dim} is not a valid dimension of {self.name}.")
        sp_dim = sp.Idx(dim)

        object.__setattr__(self, "symbolic_eq", self.symbolic_eq.subs(sp_dim, value))
        object.__setattr__(self, "_eq", self._eq.subs(sp_dim, value))
        object.__setattr__(
            self, "name", _pretty_print_dim_flags(self.name, self.dims, {dim: value})
        )
        object.__setattr__(self, "dim_vals", {dim: value})

    def to_dict(self):
        return {
            "name": self.name,
            "string_eq": self.equation,
            "symbolic_eq": self.symbolic_eq,
            "fancy_eq": self._fancy_eq,
            "standard_eq": self._eq,
            "eq_id": self.eq_id,
        }


class Result:
    def __init__(
        self,
        name,
        variables,
        parameters,
        initial_values,
        fitted_values,
        success,
        meta=None,
    ):
        self.name = name
        self.variables = variables
        self.parameters = parameters
        self.initial_values = initial_values
        self.fitted_values = fitted_values
        self.success = success
        self.fit_time = datetime.now()
        self.meta = meta if meta is not None else {}

    def to_dict(self):
        all_names = self.variables + self.parameters
        return {
            "initial": dict(zip(all_names, self.initial_values)),
            "fitted": dict(zip(all_names, self.fitted_values)),
        }

    def to_frame(self):
        return pd.DataFrame(self.to_dict())

    def format_datetime(self):
        local_timezone = time.tzname[0]
        formatted_date = self.fit_time.strftime("%A, %B %d, %Y %I:%M %p")
        return " ".join([formatted_date, local_timezone])
